###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
import re

from collections.abc import Iterator
from sys import maxsize as max_int
from typing import Any, Callable
from urllib.parse import urlparse

from everysk.utils import bool_convert

from everysk.core.datetime import DateTime, Date
from everysk.core.exceptions import ReadonlyError, FieldValueError
from everysk.core.object import BaseField


## Functions
def _do_nothing(*args, **kwargs):
    # pylint: disable=unused-argument
    raise ReadonlyError('This field value cannot be changed.')


def _min_max_validate(min_value: Any, max_value: Any, value: Any, attr_name: str) -> None:
    """
    Helper function to validate if value is between min and max for fields.

    Args:
        min_value (Any): The min value to be checked.
        max_value (Any): The max value to be checked.
        value (Any): The value used to validate.
        attr_name (str): The name of the attribute to be used on errors.

    Usage:
        >>> _min_max_validate(0, 10, 5, "my_field")
        # No exception is raised

        >>> _min_max_validate(0, 10, 15, "my_field")
        # FieldValueError is raised with the message:
        # "The value '15' for field 'my_field' must be between 0 and 10."

    Raises:
        FieldValueError: If the value is not between min and max.
    """
    if value is not None and value is not Undefined:
        min_check = False
        max_check = False

        # Check if min_value and max_value are callable
        min_value = min_value if not callable(min_value) else min_value()
        max_value = max_value if not callable(max_value) else max_value()

        if min_value is not None and min_value is not Undefined:
            min_check = value < min_value

        if max_value is not None and max_value is not Undefined:
            max_check = value > max_value

        if min_check or max_check:
            raise FieldValueError(
                f"The value '{value}' for field '{attr_name}' must be between {min_value} and {max_value}."
            )


## Classes
# For the VSCode autocomplete works correctly, we need to say what is the
# type that the method __new__ returns
class Field(BaseField):
    attr_type: Any = None

    def __new__(cls, *args, **kwargs) -> 'Field':
        return super().__new__(cls)

    def __init__(
        self,
        default: Any = None,
        required: bool = False,
        readonly: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        **kwargs
    ) -> None:
        super().__init__(attr_type=self.attr_type, default=default, required=required, readonly=readonly, required_lazy=required_lazy, empty_is_none=empty_is_none, **kwargs)


class BoolField(Field):
    attr_type: bool = bool

    def __new__(cls, *args, **kwargs) -> bool:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def clean_value(self, value: Any) -> Any:
        """
        Converts the  given value to a boolean if possible using the 'convert_boolean' function.
        (Check the function's documentation for more information)

        Args:
            value (Boolean): The value to be converted. Can be of any type

        Usage:
            >>> from everysk.core.fields import BoolField
            >>> bool_field = BoolField()
            >>> bool_field.clean_value("y")
            >>> True

            >>> bool_field.clean_value("n")
            >>> False

            >>> bool_field.clean_value("a")
            >>> ValueError: Invalid truth value 'a'

        Returns:
            Any: The value converted to its boolean corresponding
        """
        # https://docs.python.org/3/distutils/apiref.html#distutils.util.strtobool
        # The module distutils is deprecated, then we put the function code here
        if value is not None and value is not Undefined:
            value = bool_convert(value)

        return super().clean_value(value)


class ChoiceField(Field):
    attr_type: str = str
    choices: list[str] = None

    def __new__(cls, *args, **kwargs) -> str:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(
        self,
        default: str,
        choices: list,
        readonly: bool = False,
        required: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        **kwargs
    ) -> None:
        super().__init__(default=default, required=required, readonly=readonly, choices=choices, required_lazy=required_lazy, empty_is_none=empty_is_none, **kwargs)

        # Choices always need to be a list
        if not self.choices:
            self.choices = []

    def validate(self, attr_name: str, value: Any, attr_type: type = None) -> None:
        """
        Checks if value is in list of choices.
        We discard the value Undefined.

        Args:
            attr_name (str): Name of the attribute being validated
            value (Any): The value up for validation against the choices available.
            attr_type (type = None): Type of the field we are trying to validate

        Usage:
            >>> from everysk.core.fields import ChoiceField
            >>> from everysk.core.exceptions import FieldValueError
            >>> field = ChoiceField(default="option1", choices=["option1", "option2", "option3"])
            >>> try:
            >>> ... field.validate("test_field", "option1")
            >>> ... print("Validation successful")
            >>> except FieldValueError as e:
            >>> ... print(f"Validation error: {e}")
            >>> Validation successful

            >>> try:
            >>> ... field.validate("test_field", "option4")
            >>> ... print("Validation successful")
            >>> except Exception as e:
            >>> ... print(f"Validation error: {e}")
            >>> Validation error: The value 'option4' for field 'test_field' must be in this list ['option1', 'option2', 'option3'].

        Raises:
            FieldValueError: If value type doesn't match with the required type.

        Returns:
            None: This method does not return any value. It either completes successfully or raises an exception
        """
        if value is not Undefined and value not in self.choices:
            raise FieldValueError(
                f"The value '{value}' for field '{attr_name}' must be in this list {self.choices}."
            )
        super().validate(attr_name, value, attr_type)


class DateField(Field):
    attr_type: Date = Date
    min_date: Date | Callable = None
    max_date: Date | Callable = None

    def __new__(cls, *args, **kwargs) -> Date:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(
        self,
        default: Any = None,
        min_date: Date = None,
        max_date: Date = None,
        required: bool = False,
        readonly: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        **kwargs
    ) -> None:
        super().__init__(default, min_date=min_date, max_date=max_date, required=required, readonly=readonly, required_lazy=required_lazy, empty_is_none=empty_is_none, **kwargs)

    def clean_value(self, value: Any) -> Any:
        """
        This method simple converts Date strings to Date object in the Everysk format

        Args:
            value (Any): Desired date string to be converted

        Usage:
            >>> from everysk.core.fields import DateField
            >>> date_field = DateField()
            >>> date_field.clean_value("20140314")
            Date(2014, 3, 14)
            >>> date_field.clean_value("2014-03-14")
            Date(2014, 3, 14)

        Returns:
            A new Date object representing the converted date

        Raises:
            ValueError: If the string is not represented in either ISO format ("YYYY-MM-DD") or Everysk format ("YYYYMMDD").
        """
        if isinstance(value, str):
            if '-' in value:
                value = Date.fromisoformat(value)
            else:
                # Everysk format
                value = Date.strptime(value, '%Y%m%d')

        return super().clean_value(value)

    def validate(self, attr_name: str, value: Any, attr_type: type = None) -> None:
        """
        Checks if value is greater than min and lower than max including both values.

        Args:
            attr_name (str): Name of the attribute used for error checking
            value (Any): Value used for validation
            attr_type (type = None): Type of the field we are trying to validate

        Usage:
            >>> from everysk.core.fields import DateField
            >>> from everysk.core.datetime import Date
            >>> date_field = DateField(min_date=Date(2023, 1, 1), max_date=Date(2023, 12, 31))

            >>> try:
            >>> ... date_field.validate("example_date", Date(2023, 6, 15))
            >>> ... print("June 15, 2023, is a valid date")
            >>> except Exception as e:
            >>> ... print(f"Validation error: {e}")
            >>> June 15, 2023, is a valid date

            >>> try:
            >>> ... date_field.validate("example_date", Date(2022, 12, 31))
            >>> ... print("December 31, 2022, is a valid date")
            >>> except Exception as e:
            >>> ... print(f"validation error: {e}")
            >>> Validation error: The value '2022-12-31' for field 'example_date' must be between 2023-01-01 and 2023-12-31.

        Returns:
            Either the value is between min_date and max_date

        Raises:
            FieldValueError: If the value is not between min_date and max_date
        """
        _min_max_validate(self.min_date, self.max_date, value, attr_name)
        return super().validate(attr_name, value, attr_type)


class DateTimeField(Field):
    attr_type:DateTime = DateTime
    min_date: DateTime | Callable = None
    max_date: DateTime | Callable = None
    force_time: str = None

    def __new__(cls, *args, **kwargs) -> DateTime:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(
        self,
        default: Any = None,
        min_date: DateTime = None,
        max_date: DateTime = None,
        force_time: str = 'FIRST_MINUTE',
        required: bool = False,
        readonly: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        **kwargs
    ) -> None:
        super().__init__(default, min_date=min_date, max_date=max_date, force_time=force_time, required=required, readonly=readonly, required_lazy=required_lazy, empty_is_none=empty_is_none, **kwargs)

    def clean_value(self, value: Any) -> Any:
        """
        Converts a DateTime string to a DateTime object using the 'fromisoformat' function.
        (check 'fromisoformat' function for more information)

        Args:
            value (Any): desired DateTime string to be converted

        Usage:
            >>> from everysk.core.fields import DateTimeField
            >>> date_time_field = DateTimeField()

            >>> date_time_field.clean_value("2023-03-15")
            >>> DateTime(2023, 3, 15, 0, 0, tzinfo=zoneinfo.ZoneInfo(key='UTC'))

            >>> date_time_field.clean_value("2023-03-15T14:30:00")
            >>> DateTime(2023, 3, 15, 14, 30, tzinfo=zoneinfo.ZoneInfo(key='UTC'))

            >>> date_time_field.clean_value("14-03-2023")
            >>> ValueError: time data "14-03-2023" does not match format: "%Y-%m-%d"

        Returns:
            Any: a new DateTime object formatted

        Raises:
            ValueError: if the input is not formatted in the ISO format
        """
        # Convert DateTime strings to DateTime object
        if isinstance(value, str):
            if not ':' in value:
                value: DateTime = DateTime.fromisoformat(value).force_time(self.force_time)
            else:
                value: DateTime = DateTime.fromisoformat(value)
        elif Date.is_date(value):
            value: DateTime = DateTime.fromisoformat(value.isoformat())

        return super().clean_value(value)

    def validate(self, attr_name: str, value: Any, attr_type: type = None) -> None:
        """
        Checks is the value passed is in the correct range between min and max, as well as required, required_lazy, and attribute_type validations.

        Args:
            attr_name (str):
                The name of the attribute being validated. This is used for identification in error messages.

            value (Any):
                The value to be validated for the specified attribute.

            attr_type (type, optional):
                The expected type of the attribute. If provided, the value is checked to ensure it is
                of this type.

        Usage:
            >>> from everysk.core.datetime import Date
            >>> from everysk.core.fields import DateTimeField
            >>> from everysk.core.datetime import DateTime
            >>> date_time_field = DateTimeField(min_date=DateTime(2023, 1, 1), max_date=DateTime(2023, 12, 31))

            >>> try:
            >>> ... date_time_field.validate("test_field", DateTime(2023, 6, 15))
            >>> ... print("Validation successful")
            >>> except FieldValueError as e:
            >>> ... print(f"Validation error: {e}")

        Raises:
            FieldValueError: if the value provided is not between min and max values
        """
        _min_max_validate(self.min_date, self.max_date, value, attr_name)
        return super().validate(attr_name, value, attr_type)


class DictField(Field):
    attr_type: dict = dict

    class ReadonlyDict(dict):
        __setitem__ = _do_nothing
        __delitem__ = _do_nothing
        pop = _do_nothing
        popitem = _do_nothing
        clear = _do_nothing
        update = _do_nothing
        setdefault = _do_nothing

    def __new__(cls, *args, **kwargs) -> dict:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, default: Any = None, required: bool = False, readonly: bool = False, required_lazy: bool = False, empty_is_none: bool = False, **kwargs) -> None:
        # When the field is readonly, we need to change the content to.
        if readonly and default:
            default = self.ReadonlyDict(default)

        super().__init__(default, required, readonly, required_lazy, empty_is_none, **kwargs)


class EmailField(Field):
    attr_type: str = str

    def __new__(cls, *args, **kwargs) -> str:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def validate(self, attr_name: str, value: Any, attr_type: type = None) -> None:
        # The maximum length of an email is 320 characters per RFC 3696 section 3,
        # and at least 3 digits a@b
        if value is not None and value is not Undefined and isinstance(value, str) and ('@' not in value or len(value) > 320 or len(value) < 3):
            raise FieldValueError(f'Key {attr_name} must be an e-mail.')

        try:
            # Get the cases where we could have more than one @ in the string
            user, domain = value.split('@') # pylint: disable=unused-variable
        except ValueError as error:
            raise FieldValueError(f'Key {attr_name} must be an e-mail.') from error

        return super().validate(attr_name, value, attr_type)


class FloatField(Field):
    attr_type: float = float
    min_size: float = None
    max_size: float = None

    def __new__(cls, *args, **kwargs) -> float:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(self,
        default: Any = None,
        min_size: float = float('-inf'),
        max_size: float = float('inf'),
        required: bool = False,
        readonly: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        **kwargs
    ) -> None:
        super().__init__(default, required, readonly, min_size=min_size, max_size=max_size, required_lazy=required_lazy, empty_is_none=empty_is_none, **kwargs)

    def clean_value(self, value: Any) -> Any:
        """
        Convert a float string to a float object

        Args:
            value (Any): Value desired for conversion to float

        Usage:
            >>> from everysk.core.fields import FloatField
            >>> float_field = FloatField()

            >>> float_field.clean_value("3.14")
            >>> 3.14

            >>> float_field.clean_value(123)
            >>> 123.0

            >>> float_field.clean_value('abc')
            >>> ValueError: could not convert string to float: 'abc'

        Returns:
            Float: if value is able to be converted to float object

        Raises:
            ValueError: If value could not be converted to a float
        """
        # Convert Float strings to float object
        if isinstance(value, (int, str)):
            value = float(value)

        return super().clean_value(value)

    def validate(self, attr_name: str, value: Any, attr_type: type = None) -> None:
        """
        Checks if value is greater than min and lower than max including both values.
        It also validates the `required`, `required_lazy`, and `attr_type` parameters according to the `_validate()` function.

        Raises:
            FieldValueError: If value is not between min and max.

        Usage:
            >>> from everysk.core.fields import FloatField
            >>> from everysk.core.exceptions import FieldValueError
            >>> float_field = FloatField(min_size=0.0, max_size=100.0)

            >>> try:
            >>> ... float_field.validate("test_field", 50.0)
            >>> ... print("Validation successful: 50.0 is within the range.")
            >>>  except FieldValueError as e:
            >>> ... print(f"Validation error: {e}")
            >>> Validation successful: 50.0 is within the range.

            >>> try:
            >>> ... float_field.validate("test_field", -10.0)
            >>> ... print("Validation successful: -10.0 is within the range.")
            >>>  except FieldValueError as e:
            >>> ... print(f"Validation error: {e}")
            >>> Validation error: The value '-10.0' for field 'test_field' is not within the allowed range.
        """
        _min_max_validate(min_value=self.min_size, max_value=self.max_size, value=value, attr_name=attr_name)
        return super().validate(attr_name, value, attr_type)


class IntField(Field):
    attr_type: int = int
    min_size: int = None
    max_size: int = None

    def __new__(cls, *args, **kwargs) -> int:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(self,
        default: Any = None,
        min_size: int = -max_int,
        max_size: int = max_int,
        required: bool = False,
        readonly: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        **kwargs
    ) -> None:
        super().__init__(default, required, readonly, min_size=min_size, max_size=max_size, required_lazy=required_lazy, empty_is_none=empty_is_none, **kwargs)

    def clean_value(self, value: Any) -> Any:
        """
        Converts a value to an integer if it can be converted

        Args:
            value (Any): Value to convert


        Usage:
            >>> from everysk.core.fields import IntField
            >>> int_field = IntField()

            >>> int_field.clean_value("123")
            >>> 123

            >>> int_field.clean_value("abc")
            >>> ValueError: invalid literal for int() with base 10: 'abc'

        Returns:
            Int: value converted to an integer

        Raises:
            ValueError: if the string cannot be converted to an integer
        """
        # Convert Int strings to int object
        if isinstance(value, str):
            value = int(value)

        return super().clean_value(value)

    def validate(self, attr_name: str, value: Any, attr_type: type = None) -> None:
        """
        Checks if value is greater than min and lower than max including both values.
        It also validates the `required`, `required_lazy`, and `attr_type` parameters as per the `_validate()` function.

        Args:
            attr_name (str):
                The name of the attribute being validated. This name is used in the error message if the
                validation fails.

            value (Any):
                The value to be validated. The function checks if this value is within the defined min and max range.

            attr_type (type, optional):
                The type against which the value will be validated. If provided, the method
                ensures that the value is of this type.

        Usage:
            >>> from everysk.core.fields import IntField
            >>> from everysk.core.exceptions import FieldValueError
            >>> int_field = IntField(min_size=1, max_size=10)

            >>> try:
            >>> ... int_field.validate("test_field", 5)
            >>> ... print("Validation successful")
            >>> except FieldValueError as e:
            >>> ... print(f"Validation error: {e}")
            >>> Validation successful

            >>> try:
            >>> ... int_field.validate("test_field", -1)
            >>> ... print("validation successful")
            >>> except FieldValueError as e:
            >>> ...print(f"Validation error: {e}")
            >>> Validation error: The value '-1' for field 'test_field' must be between 1 and 10.

        Raises:
            FieldValueError: If value is not between min_size and max.
        """
        _min_max_validate(min_value=self.min_size, max_value=self.max_size, value=value, attr_name=attr_name)
        return super().validate(attr_name, value, attr_type)


class IteratorField(Field):
    attr_type: Iterator = Iterator

    def __new__(cls, *args, **kwargs) -> Iterator:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def clean_value(self, value: Any) -> Any:
        """
        Cleans the given value before storing it.

        Args:
            value (Any): The value to be cleaned.

        Returns:
            Any: The cleaned value.

        """
        # Convert List/Str to iterators
        if isinstance(value, (list, str)):
            value = iter(value)

        return super().clean_value(value)


# https://stackoverflow.com/questions/855191/how-big-can-a-python-list-get#comment112727918_15739630
# so on 64 systems is maxsize / 8
class ListField(Field):
    attr_type: list = list
    min_size: int = None
    max_size: int = None

    class ReadonlyList(list):
        __setitem__ = _do_nothing
        __delitem__ = _do_nothing
        append = _do_nothing
        clear = _do_nothing
        extend = _do_nothing
        insert = _do_nothing
        pop = _do_nothing
        remove = _do_nothing
        reverse = _do_nothing
        sort = _do_nothing

    def __new__(cls, *args, **kwargs) -> list:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(
        self,
        default: Any = None,
        min_size: int = 0,
        max_size: int = max_int / 8,
        readonly: bool = False,
        required: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        **kwargs
    ) -> None:
        if min_size < 0:
            raise FieldValueError('List min_size cloud not be a negative number.')

        # When the field is readonly, we need to change the content to.
        if readonly and (default is not None or default is not Undefined):
            default = self.ReadonlyList(default)

        super().__init__(default=default, min_size=min_size, max_size=max_size, readonly=readonly, required=required, required_lazy=required_lazy, empty_is_none=empty_is_none, **kwargs)

    def clean_value(self, value: Any) -> Any:
        """
        Clean the value before storing it.
        This method is used to ensure that the value is in the correct format before storing it.
        This method checks if the value is a string and converts it into a list if necessary.
        It then calls the parent class's clean_value method to perform additional cleaning.

        Args:
            value (Any): The value to be cleaned.

        Example Usage:
            >>> field = Field()
            >>> field.clean_value("example")
            ['example']

        Returns:
            Any: The cleaned value.
        """
        if isinstance(value, str):
            value = [value]

        return super().clean_value(value)

    def validate(self, attr_name: str, value: Any, attr_type: type = None) -> None:
        """
        Checks if value is greater than min_size and lower than max_size including both values.

        Raises:
            FieldValueError: If value is not list instance or if is not between min_size and max_size
        """
        if value is not None and value is not Undefined:
            if not isinstance(value, list):
                raise FieldValueError(f"The '{attr_name}' value must be a list.")
            if not self.min_size <= len(value) <= self.max_size:
                raise FieldValueError(
                    f"The attribute '{attr_name}' is not within the specified list range. min_size: {self.min_size} max_size: {self.max_size}"
                )
        super().validate(attr_name, value, attr_type)


class SetField(Field):
    attr_type: set = set
    min_size: int = None
    max_size: int = None

    class ReadonlySet(set):
        __setitem__ = _do_nothing
        __delitem__ = _do_nothing
        add = _do_nothing
        clear = _do_nothing
        discard = _do_nothing
        pop = _do_nothing
        remove = _do_nothing
        update = _do_nothing

    def __new__(cls, *args, **kwargs) -> set:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(
        self,
        default: Any = None,
        min_size: int = 0,
        max_size: int = max_int / 8,
        readonly: bool = False,
        required: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        **kwargs
    ) -> None:
        if min_size < 0:
            raise FieldValueError('Set min_size cloud not be a negative number.')

        # When the field is readonly, we need to change the content to.
        if readonly and (default is not None or default is not Undefined):
            default = self.ReadonlySet(default)

        super().__init__(default=default, min_size=min_size, max_size=max_size, readonly=readonly, required=required, required_lazy=required_lazy, empty_is_none=empty_is_none, **kwargs)

    def validate(self, attr_name: str, value: Any, attr_type: type = None) -> None:
        """
        Checks if value is greater than min_size and lower than max_size including both values.

        Raises:
            FieldValueError: If value is not list instance.
            FieldValueError: If value is not between min_size and max_size.
        """
        if value is not None and value is not Undefined and isinstance(value, set):
            if not self.min_size <= len(value) <= self.max_size:
                raise FieldValueError(
                    f"The attribute '{attr_name}' is not within the specified set range. min_size: {self.min_size} max_size: {self.max_size}"
                )
        super().validate(attr_name, value, attr_type)


class StrField(Field):
    """
    Represents a string field in an entity.
    This field type is used to store string values and provides validation
    for string attributes based on specified constraints such as minimum and maximum size
    and regular expression pattern matching.

    Args:
        default (Any, optional):
            The default value for the field. Defaults to None.
        min_size (int, optional):
            The minimum allowed length for the string. Defaults to 0.
        max_size (int, optional):
            The maximum allowed length for the string. Defaults to the maximum integer value.
        regex (str, optional):
            The regular expression pattern to match against the string value. Defaults to None.
        readonly (bool, optional):
            Whether the field is read-only. Defaults to False.
        required (bool, optional):
            Whether the field is required. Defaults to False.
        required_lazy (bool, optional):
            Whether the field is lazily required. Defaults to False.
        empty_is_none (bool, optional):
            Whether an empty string should be treated as None. Defaults to False.

    Raises:
        FieldValueError: If the minimum size is negative
    """

    attr_type: str = str
    min_size: int = None
    max_size: int = None
    regex: re.Pattern = None

    def __new__(cls, *args, **kwargs) -> str:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(
        self,
        default: Any = None,
        min_size: int = 0,
        max_size: int = max_int,
        regex: str = None,
        readonly: bool = False,
        required: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        **kwargs
    ) -> None:
        if min_size < 0:
            raise FieldValueError('String min_size cloud not be a negative number.')
        super().__init__(default=default, regex=regex, min_size=min_size, max_size=max_size, readonly=readonly, required=required, required_lazy=required_lazy, empty_is_none=empty_is_none, **kwargs)

    def validate(self, attr_name: str, value: Any, attr_type: type = None) -> None:
        """
        Validates a string attribute against specified constraints.

        Args:
            attr_name (str):
                The name of the attribute being validated.
            value (Any):
                The value of the attribute to be validated.
            attr_type (type, optional):
                The expected type of the attribute. Defaults to None.

        Raises:
            FieldValueError:
                If the provided value does not meet the specified constraints.
                If the value does not match the regular expression pattern.
                If the length of the value is outside the specified range.
        """
        # First let all default checks be done.
        super().validate(attr_name, value, attr_type)

        if self.regex and value and value is not Undefined and not self.regex.match(value):
            raise FieldValueError(
                f"The value '{value}' for field '{attr_name}' must match with this regex: {self.regex.pattern}."
            )

        # Then we check for the size.
        try:
            if value is not None and value is not Undefined:
                _min_max_validate(min_value=self.min_size, max_value=self.max_size, value=len(value), attr_name=attr_name)
        except FieldValueError as error:
            raise FieldValueError(
                f"The length '{len(value)}' for attribute '{attr_name}' must be between '{self.min_size}' and '{self.max_size}'."
            ) from error


class RegexField(Field):
    attr_type: re.Pattern = re.Pattern

    def __new__(cls, *args, **kwargs) -> re.Pattern:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def __init__(
        self,
        default: Any = None,
        readonly: bool = False,
        required: bool = False,
        required_lazy: bool = False,
        empty_is_none: bool = False,
        **kwargs
    ) -> None:
        super().__init__(default=default, readonly=readonly, required=required, required_lazy=required_lazy, empty_is_none=empty_is_none, **kwargs)


    def clean_value(self, value: Any) -> Any:
        """
        Clean the value by compiling it into a regular expression pattern.

        This method takes a value and compiles it into a regular expression pattern if it is a string.
        If the value is equal to the default value, it updates the default value with the compiled pattern.
        It then calls the parent class's clean_value method to perform any additional cleaning.

        Args:
            value (Any): The value to be cleaned.

        Example Usage:
            >>> field = RegexField(default='[0-9]+')
            >>> field.clean_value('123')
            re.compile('[0-9]+')

        Returns:
            Any: The cleaned value.
        """
        if isinstance(value, str):
            value = re.compile(value)
        return super().clean_value(value)


class TupleField(Field):
    attr_type: tuple = tuple

    def __new__(cls, *args, **kwargs) -> tuple:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)


class URLField(Field):
    attr_type: str = str
    supported_schemes: set = {'http', 'https', 'ftp', 'ftps', 'git'}

    def __new__(cls, *args, **kwargs) -> str:
        # This signature is used to change the default autocomplete for this class
        return super().__new__(cls, *args, **kwargs)

    def validate(self, attr_name: str, value: Any, attr_type: type = None) -> None:
        if isinstance(value, str):
            # https://github.com/django/django/blob/main/django/core/validators.py
            ul = "\u00a1-\uffff"  # Unicode letters range (must not be a raw string).

            # IP patterns
            ipv4_re = (
                r"(?:0|25[0-5]|2[0-4][0-9]|1[0-9]?[0-9]?|[1-9][0-9]?)"
                r"(?:\.(?:0|25[0-5]|2[0-4][0-9]|1[0-9]?[0-9]?|[1-9][0-9]?)){3}"
            )
            ipv6_re = r"\[[0-9a-f:.]+\]"  # (simple regex, validated later)

            # Host patterns
            hostname_re = (
                r"[a-z" + ul + r"0-9](?:[a-z" + ul + r"0-9-]{0,61}[a-z" + ul + r"0-9])?"
            )

            # Max length for domain name labels is 63 characters per RFC 1034 sec. 3.1
            domain_re = r"(?:\.(?!-)[a-z" + ul + r"0-9-]{1,63}(?<!-))*"
            tld_re = (
                r"\."  # dot
                r"(?!-)"  # can't start with a dash
                r"(?:[a-z" + ul + "-]{2,63}"  # domain label
                r"|xn--[a-z0-9]{1,59})"  # or punycode label
                r"(?<!-)"  # can't end with a dash
                r"\.?"  # may have a trailing dot
            )
            host_re = "(" + hostname_re + domain_re + tld_re + "|localhost)"

            regex = re.compile(
                r"^(?:[a-z0-9.+-]*)://"  # scheme is validated separately
                r"(?:[^\s:@/]+(?::[^\s:@/]*)?@)?"  # user:pass authentication
                r"(?:" + ipv4_re + "|" + ipv6_re + "|" + host_re + ")"
                r"(?::[0-9]{1,5})?"  # port
                r"(?:[/?#][^\s]*)?"  # resource path
                r"\Z",
                re.IGNORECASE,
            )
            if not regex.match(value):
                raise FieldValueError(f'Key {attr_name} must be an URL.')

            # The scheme needs a separate check
            try:
                result = urlparse(value)
                if result.scheme not in self.supported_schemes:
                    raise FieldValueError(f'Key {attr_name} must be an URL.')

            except ValueError:
                raise FieldValueError(f'Key {attr_name} must be an URL.') from ValueError

        return super().validate(attr_name, value, attr_type)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
import logging
import sys
from contextlib import AbstractContextManager
from contextvars import ContextVar
from types import TracebackType
from threading import Thread
from typing import Any
from everysk.config import settings


# This only makes sense in Production
try:
    # Imports the Cloud Logging client library
    from google.cloud.logging.handlers import StructuredLogHandler # type: ignore
except ImportError:
    StructuredLogHandler = None


class LoggerManager(AbstractContextManager):
    labels: ContextVar = ContextVar('everysk-log-labels', default={})
    token: str = None

    def __init__(self, labels: dict) -> None:
        """
        Context class to create a context manager for the Logger object.

        Args:
            labels (dict): The extra labels sent to Google Logging.

        Example:

            >>> from everysk.core.log import Logger, LoggerManager
            >>> log = Logger(name='log-test')
            >>> log.debug('Test')
            2024-02-07 12:49:10,640 - DEBUG - {} - Test
            >>> with LoggerManager(labels={'user_id': 1}):
            ...     log.debug('Test')
            2024-02-07 12:52:07,319 - DEBUG - {'user_id': 1} - Test
        """
        self.token = self.labels.set(labels)

    def __exit__(self, __exc_type: type[BaseException] | None, __exc_value: BaseException | None, __traceback: TracebackType | None) -> bool | None:
        """
        https://docs.python.org/3/library/stdtypes.html#contextmanager.__exit__

        Returns:
            bool | None: If return is False any exception will be raised.
        """
        self.labels.reset(self.token)
        return False


# The Logger cannot be a BaseObject class because it could be used inside base things
class Logger:
    ## Private attributes
    # This needs to be initialized with an empty set because it will be a global list
    _deprecated_hash: set[str] = set()

    ## Public attributes
    google_project: str = None
    labels: dict = None
    level: int = None
    log: logging.Logger = None
    name: str = None
    output_format: str = None
    slack_url: str = None
    stacklevel: int = None

    def __init__(self, name: str = Undefined, google_project: str = Undefined, labels: dict = Undefined, level: int = Undefined, output_format: str = Undefined, slack_url: str = Undefined, stacklevel: int = Undefined) -> None:
        """
        Logger class used to send messages to STDOUT or Google CLoud Logging.

        Args:
            name (str, optional): The name of the log. Defaults to "root".
            google_project (str, optional): The name of the project inside Google. Defaults to Undefined.
            labels (dict, optional): Extra information sent to the log. Defaults to {}.
            level (int, optional): The level of the messages that this log will handle. Defaults to DEBUG.
            output_format (str, optional): The string that represents the output format for the log. Defaults to "%(asctime)s - %(levelname)s - %(labels)s - %(message)s".
            stacklevel (int, optional): The position used to show what file generated the message. Defaults to 2.

        Example:

            >>> from everysk.core.log import Logger
            >>> log = Logger(name='log-test')
            >>> log.debug('Test')
            2024-02-07 12:49:10,640 - DEBUG - {} - Test
        """
        if name == 'root':
            raise ValueError('The name of the log could not be "root".')

        self._set_value(name='name', value=name, default='everysk-root-log', klass=str)
        self._set_value(name='google_project', value=google_project, default=settings.EVERYSK_GOOGLE_CLOUD_PROJECT, klass=str)
        self._set_value(name='labels', value=labels, default={}, klass=dict)
        self._set_value(name='level', value=level, default=logging.DEBUG, klass=int)
        self._set_value(name='output_format', value=output_format, default='%(asctime)s - %(levelname)s - %(labels)s - %(message)s', klass=str)
        self._set_value(name='slack_url', value=slack_url, default=None, klass=str)
        self._set_value(name='stacklevel', value=stacklevel, default=2, klass=int)
        self.log = self.get_python_logger()

    ## Private methods
    def _set_value(self, name: str, value: Any, default: Any, klass: type) -> None:
        """
        Method used to validate the type and set the value.

        Args:
            name (str): Attribute name.
            value (Any): Value that should be validated.
            default (Any): Default value if "value" is Undefined.
            klass (type): Class used to check if the "value" is an instance.

        Raises:
            ValueError: If the value is not of the klass type.

        Returns:
            None

        Usage:
            >>> from everysk.core.log import Logger
            >>> obj = Logger()
            >>> obj._set_value(name='example_attribute', value=42, default=0, klass=int)
            >>> print(obj.example_attribute)
            >>> 42

        """
        if value is Undefined:
            value = default
        elif not isinstance(value, klass):
            raise ValueError(f'The "{name}" attribute must be a {klass}.')

        setattr(self, name, value)

    ## Methods
    def get_python_logger(self) -> logging.Logger:
        """
        Method that creates/get the Python Logger object and attach the correct handler.
        If you need to deactivate the Google Logging Handler, set the settings.EVERYSK_GOOGLE_CLOUD_LOGGING_INTEGRATION
        to False or do not install the Handler pip package.
        The default handler will be the stdout.

        Returns:
            logging.Logger: The Python Logger object.

        Usage:
            >>> logger = Logger(name='example_log')
            >>> python_logger = logger.get_python_logger()
            >>> print(python_logger)
            <Logger example_log (DEBUG)>

        """
        # Create the log
        log = logging.getLogger(self.name)
        log.setLevel(self.level)
        log.propagate = False # Don't pass message to others loggers

        # We should only have one handler per log name
        if not hasattr(log, 'handler'):
            if settings.EVERYSK_GOOGLE_CLOUD_LOGGING_INTEGRATION and self.google_project and StructuredLogHandler:
                log.handler = StructuredLogHandler(project_id=self.google_project, labels=self.labels)
            else:
                log.handler = logging.StreamHandler(stream=sys.stdout)

                # Set the format that the message is displayed
                log.handler.setFormatter(logging.Formatter(self.output_format))

            # Add the handler inside the log
            log.addHandler(log.handler)

        # Set the level that the handler will be using.
        log.handler.setLevel(self.level)

        return log

    def get_labels(self) -> dict:
        """
        The labels are extra information that the Google Logging uses to create filters.

        Normally we set one at every request to be able to filter all logs from one request.
        If they are not set an empty dict will be the default value.

        Returns:
            dict: The labels dictionary

        Usage:
            >>> from everysk.core.log import Logger
            >>> logger = Logger()
            >>> labels = logger.get_labels()
        """
        labels = self.labels.copy()
        context_labels = LoggerManager.labels.get()
        if context_labels:
            labels.update(context_labels)

        return labels

    def _get_kwargs(self) -> dict:
        """
        Method that create a dict to be used as **kwargs for all logs.
        """
        return {
            'extra': {'labels': self.get_labels()}, # Set extra labels for Google
            'stacklevel': self.stacklevel, # Change the source to be one of the "parents" caller
        }

    def _show_deprecated(self, _id: str, show_once: bool) -> bool:
        """
        If show_once is False this always return True, otherwise
        checks if this _id is in the self._deprecated_hash set.

        Args:
            _id (str): String that will be stored to be checked later
            show_once (bool): Flag to store or not the id.
        """
        if show_once:
            if _id in self._deprecated_hash:
                return False

            self._deprecated_hash.add(_id)

        return True

    ## Helpers
    def critical(self, msg: str, *args) -> None:
        """ Log 'msg % args' with severity 'CRITICAL'. """
        self.log.critical(msg, *args, **self._get_kwargs())

    def debug(self, msg: str, *args) -> None:
        """ Log 'msg % args' with severity 'DEBUG'. """
        self.log.debug(msg, *args, **self._get_kwargs())

    def error(self, msg: str, *args) -> None:
        """ Log 'msg % args' with severity 'ERROR'. """
        self.log.error(msg, *args, **self._get_kwargs())

    def exception(self, msg: str, *args) -> None:
        """ Log 'msg % args' with severity 'ERROR' with exception information. """
        self.log.exception(msg, *args, **self._get_kwargs())

    def info(self, msg: str, *args) -> None:
        """ Log 'msg % args' with severity 'INFO'. """
        self.log.info(msg, *args, **self._get_kwargs())

    def warning(self, msg: str, *args) -> None:
        """ Log 'msg % args' with severity 'WARNING'. """
        self.log.warning(msg, *args, **self._get_kwargs())

    def deprecated(self, msg: str, *args, show_once: bool = True) -> None:
        """
        Shows a DeprecationWarning message with severity 'WARNING'.
        If show_once is True, then the message will be showed only once.

        Args:
            msg (str): The message that must be showed.
            show_once (bool, optional): If the message must be showed only once. Defaults to True.
        """
        _id = hash(f'{msg}, {args}')
        if self._show_deprecated(_id=_id, show_once=show_once):
            msg = f'DeprecationWarning: {msg}'
            self.warning(msg, *args)

    def slack(self, title: str, message: str, color: str, url: str = None) -> None:
        """
        Send a message to a Slack channel using Slack WebHooks.
        https://api.slack.com/messaging/webhooks
        The same message will be sent to the default log:
            danger -> error
            success -> info
            warning -> warning

        Args:
            title (str): The title of the message.
            message (str): The body of the message.
            color (str): 'danger' | 'success' | 'warning'
            url (str, optional): The slack webhook url. Defaults to self.slack_url.
        """
        if url is None:
            if self.slack_url is None:
                url = settings.SLACK_URL
            else:
                url = self.slack_url

        # We send the message only if url is set and we are in PROD
        if url and settings.PROFILE == 'PROD' and 'unittest' not in sys.modules:
            # The import must be here to avoid circular import inside http module
            from everysk.core.slack import Slack # pylint: disable=import-outside-toplevel
            client = Slack(title=title, message=message, color=color, url=url)
            # This will send the message to Slack without block the request
            Thread(target=client.send).start()

        log_message = f'Slack message: {title} -> {message}'
        if color == 'danger':
            self.error(log_message)

        elif color == 'success':
            self.info(log_message)

        elif color == 'warning':
            self.warning(log_message)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# pylint: disable=protected-access, attribute-defined-outside-init
from copy import copy, deepcopy
from typing import Any, Self
from unittest import TestCase
from everysk.core.datetime import DateTime, Date
from everysk.core.exceptions import DefaultError, FieldValueError, RequiredError
from everysk.core.fields import BoolField
from everysk.core.object import BaseField, BaseDict, BaseObject, _required, _validate, MetaClass


class RequiredTestCase(TestCase):

    def test_required(self):
        self.assertRaises(RequiredError, _required, attr_name='Test', value=None)
        self.assertRaises(RequiredError, _required, attr_name='Test', value='')
        self.assertRaises(RequiredError, _required, attr_name='Test', value=())
        self.assertRaises(RequiredError, _required, attr_name='Test', value=[])
        self.assertRaises(RequiredError, _required, attr_name='Test', value={})


class ValidateTestCase(TestCase):

    def test_none(self):
        _validate(attr_name='test', value=None, attr_type=Date)
        _validate(attr_name='test', value=None, attr_type=DateTime)
        _validate(attr_name='test', value=None, attr_type=dict)
        _validate(attr_name='test', value=None, attr_type=float)
        _validate(attr_name='test', value=None, attr_type=int)
        _validate(attr_name='test', value=None, attr_type=list)
        _validate(attr_name='test', value=None, attr_type=str)

    def test_undefined(self):
        _validate(attr_name='test', value=Undefined, attr_type=Date)
        _validate(attr_name='test', value=Undefined, attr_type=DateTime)
        _validate(attr_name='test', value=Undefined, attr_type=dict)
        _validate(attr_name='test', value=Undefined, attr_type=float)
        _validate(attr_name='test', value=Undefined, attr_type=int)
        _validate(attr_name='test', value=Undefined, attr_type=list)
        _validate(attr_name='test', value=Undefined, attr_type=str)

    def test_date(self):
        _validate(attr_name='test', value=Date.today(), attr_type=Date)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=DateTime.now(), attr_type=Date)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value='1', attr_type=Date)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1, attr_type=Date)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1.0, attr_type=Date)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=[], attr_type=Date)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value={}, attr_type=Date)

    def test_datetime(self):
        _validate(attr_name='test', value=DateTime.now(), attr_type=DateTime)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=Date.today(), attr_type=DateTime)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value='1', attr_type=DateTime)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1, attr_type=DateTime)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1.0, attr_type=DateTime)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=[], attr_type=DateTime)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value={}, attr_type=DateTime)

    def test_dict(self):
        _validate(attr_name='test', value={}, attr_type=dict)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=DateTime.now(), attr_type=dict)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value='1', attr_type=dict)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1, attr_type=dict)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1.0, attr_type=dict)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=[], attr_type=dict)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=Date.today(), attr_type=dict)

    def test_float(self):
        _validate(attr_name='test', value=1.0, attr_type=float)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=DateTime.now(), attr_type=float)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value='1', attr_type=float)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1, attr_type=float)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=Date.today(), attr_type=float)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=[], attr_type=float)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value={}, attr_type=float)

    def test_int(self):
        _validate(attr_name='test', value=1, attr_type=int)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=DateTime.now(), attr_type=int)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value='1', attr_type=int)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=Date.today(), attr_type=int)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1.0, attr_type=int)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=[], attr_type=int)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value={}, attr_type=int)

    def test_list(self):
        _validate(attr_name='test', value=[], attr_type=list)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=DateTime.now(), attr_type=list)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value='1', attr_type=list)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1, attr_type=list)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1.0, attr_type=list)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=Date.today(), attr_type=list)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value={}, attr_type=list)

    def test_str(self):
        _validate(attr_name='test', value='1', attr_type=str)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=DateTime.now(), attr_type=str)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=Date.today(), attr_type=str)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1, attr_type=str)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=1.0, attr_type=str)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value=[], attr_type=str)
        self.assertRaises(FieldValueError, _validate, attr_name='test', value={}, attr_type=str)

    def test_any(self):
        _validate(attr_name='test', value=None, attr_type=Any)
        _validate(attr_name='test', value={}, attr_type=Any)
        _validate(attr_name='test', value=[], attr_type=Any)
        _validate(attr_name='test', value=123, attr_type=Any)
        _validate(attr_name='test', value='123', attr_type=Any)
        _validate(attr_name='test', value=callable, attr_type=Any)
        _validate(attr_name='test', value=Undefined, attr_type=Any)
        _validate(attr_name='test', value=Undefined.default_parse_string, attr_type=Any)

    def test_callable(self):
        _validate(attr_name='test', value=lambda x: x, attr_type=callable)
        with self.assertRaisesRegex(FieldValueError, 'Key test must be <built-in function callable>.'):
            _validate(attr_name='test', value='func', attr_type=callable)

    def test_annotation_string(self):
        _validate(attr_name='test', value=object(), attr_type='object')
        with self.assertRaisesRegex(FieldValueError, 'Key test must be object.'):
            _validate(attr_name='test', value=str(), attr_type='object')


class BaseFieldTestCase(TestCase):

    def test_init(self):
        field = BaseField(attr_type=str, default='default', required=True, readonly=True, other='Test')
        self.assertEqual(field.attr_type, str)
        self.assertEqual(field.default, 'default')
        self.assertTrue(field.required)
        self.assertTrue(field.readonly)
        self.assertEqual(field.other, 'Test') # pylint: disable=no-member

    def test_init_default_empty(self):
        self.assertRaises(DefaultError, BaseField, attr_type=str, default=[])
        self.assertRaises(DefaultError, BaseField, attr_type=str, default={})

    def test_init_default_not_empty(self):
        BaseField(attr_type=str, default=[1, 2, 3])
        BaseField(attr_type=str, default={'a': 1, 'b': 2})

    def test_init_readonly_default(self):
        self.assertRaises(RequiredError, BaseField, attr_type=str, readonly=True)

    def test_clean(self):
        field = BaseField(attr_type=str)
        self.assertEqual(field.clean_value('Test'), 'Test')

    def test_transform_to_none(self):
        field = BaseField(attr_type=str, empty_is_none=False)
        self.assertEqual(field.transform_to_none(''), '')
        field = BaseField(attr_type=str, empty_is_none=True)
        self.assertEqual(field.transform_to_none(''), None)

    def test_validate_attr_type(self):
        field = BaseField(attr_type=str)
        self.assertRaises(FieldValueError, field.validate, attr_name='test', value=True)
        self.assertRaises(FieldValueError, field.validate, attr_name='test', value='1', attr_type=int)

    def test_validate_readonly(self):
        field = BaseField(attr_type=str, readonly=True, default='Teste')
        self.assertRaises(FieldValueError, field.validate, attr_name='test', value='new')

    def test_validate_required(self):
        field = BaseField(attr_type=str, required=True)
        self.assertRaises(RequiredError, field.validate, attr_name='test', value=None)
        field = BaseField(default=Undefined, attr_type=int, required=True)
        self.assertRaises(RequiredError, field.validate, attr_name='test', value=None)
        field = BaseField(default=Undefined.default_parse_string, attr_type=int, required=True)
        self.assertRaises(RequiredError, field.validate, attr_name='test', value=None)

    def test_validate_required_lazy(self):
        field = BaseField(attr_type=str, required=False, required_lazy=True)
        field.validate(attr_name='test', value=None)
        field = BaseField(attr_type=int, required=False, required_lazy=True)
        field.validate(attr_name='test', value=Undefined)

    def test_repr(self):
        field = BaseField(attr_type=str)
        self.assertEqual(repr(field), 'BaseField')

    def test_get_value(self):
        def func():
            return 'Test'
        field = BaseField(attr_type=callable)
        self.assertEqual(field.get_value(func), 'Test')

    def test_required_and_required_lazy_error(self):
        self.assertRaisesRegex(
            FieldValueError,
            "Required and required_lazy can't be booth True.",
            BaseField,
            attr_type=str,
            required=True,
            required_lazy=True
        )

    def test_equal(self):
        self.assertEqual(
            BaseField(attr_type=str, default='field'),
            BaseField(attr_type=str, default='field')
        )

    def test_not_equal(self):
        self.assertNotEqual(
            BaseField(attr_type=str, default='field1'),
            BaseField(attr_type=str, default='field2')
        )

    def test_getattr_method(self):
        field = BaseField(attr_type=str)
        self.assertTrue(hasattr(field, 'startswith'))
        # int.real
        self.assertFalse(hasattr(field, 'real'))


## Fake class for BaseObject tests.
class BaseObjectTestClass(BaseObject):
    attr: int = None

class BaseObjectFieldsTestClass(BaseObject):
    _private: str = 'Private'
    normal_attr = 'Normal'
    field1 = BaseField(attr_type=int)
    field_default = BaseField(attr_type=int, default=100)
    field_readonly = BaseField(attr_type=str, default='default_value', readonly=True)

class BaseObjectAttributeInheritanceTestClass(BaseObjectFieldsTestClass):
    local_attr = BaseField(attr_type=str, default='Local Field')

class BaseObjectInheritanceTestClass(BaseObjectTestClass):
    attr_other: str = None

class BaseObjectRequiredTestClass(BaseObject):
    required = BaseField(attr_type=int, required=True)
    required1 = BaseField(attr_type=int, required=True)

class BaseObjectRequiredInheritanceTestClass(BaseObjectRequiredTestClass):
    required1 = 1

class BaseObjectRequiredLazyTestClass(BaseObject):
    required = BaseField(attr_type=int, required_lazy=True)
    required1 = BaseField(attr_type=int, required_lazy=True)

class BaseObjectRequiredLazyInheritanceTestClass(BaseObjectRequiredLazyTestClass):
    required1 = 1

class BaseObjectAttributeReplace(BaseObject):
    is_parent = BaseField(attr_type=bool, default=True)

class BaseObjectAttributeReplaceChild(BaseObjectAttributeReplace):
    is_parent = False

class BaseFieldClean(BaseField):

    def clean_value(self, value: Any) -> Any:
        if value not in (None, int):
            raise FieldValueError('Error')

class BaseObjectCleanValueError(BaseObject):
    field1 = BaseFieldClean(attr_type=int)

class BaseObjectTestCase(TestCase):

    def test_init(self):
        obj = BaseObjectTestClass()
        self.assertIsNone(obj.attr)

    def test_attr_creation(self):
        obj = BaseObjectTestClass()
        obj.attr = 1
        self.assertEqual(obj.attr, 1)

    def test_fields(self):
        obj = BaseObjectFieldsTestClass()
        self.assertEqual(obj._private, 'Private')
        self.assertEqual(obj.normal_attr, 'Normal')
        self.assertIsNone(obj.field1)
        self.assertEqual(obj.field_default, 100)
        self.assertEqual(obj.field_readonly, 'default_value')
        self.assertRaises(FieldValueError, BaseObjectFieldsTestClass, field_readonly='Changed')
        with self.assertRaisesRegex(FieldValueError, "The field 'field_readonly' value cannot be changed."):
            obj.field_readonly = 'Changed'

    def test_required(self):
        self.assertRaises(RequiredError, BaseObjectRequiredTestClass)

    def test_required_inheritance(self):
        self.assertRaisesRegex(RequiredError, 'The required attribute is required.', BaseObjectRequiredInheritanceTestClass)
        obj = BaseObjectRequiredInheritanceTestClass(required=2)
        self.assertEqual(obj.required, 2)
        self.assertEqual(obj.required1, 1)

    def test_required_lazy(self):
        obj = BaseObjectRequiredLazyTestClass()
        self.assertRaisesRegex(
            RequiredError,
            'The required attribute is required.',
            obj.validate_required_fields
        )

    def test_required_lazy_inheritance(self):
        obj = BaseObjectRequiredLazyInheritanceTestClass()
        self.assertRaisesRegex(
            RequiredError,
            'The required attribute is required.',
            obj.validate_required_fields
        )
        obj.required = 2
        obj.validate_required_fields()
        self.assertEqual(obj.required, 2)
        self.assertEqual(obj.required1, 1)

    def test_default_receive_none(self):
        obj = BaseObjectFieldsTestClass(field_default=None)
        self.assertEqual(obj.field_default, None)

    def test_attribute_change_bool_type(self):
        parent = BaseObjectAttributeReplace()
        self.assertTrue(parent.is_parent)
        child = BaseObjectAttributeReplaceChild()
        self.assertFalse(child.is_parent)

    def test_get_attribute_class(self):
        self.assertEqual(BaseObjectFieldsTestClass.field_default, 100)

    def test_get_attribute_inheritance_class(self):
        self.assertEqual(BaseObjectAttributeInheritanceTestClass.local_attr, 'Local Field')
        self.assertEqual(BaseObjectAttributeInheritanceTestClass.field_readonly, 'default_value')

    def test_get_attribute_class_error(self):
        with self.assertRaisesRegex(AttributeError, "type object 'BaseObjectFieldsTestClass' has no attribute 'field_default_erro'"):
            BaseObjectFieldsTestClass.field_default_erro # pylint: disable=pointless-statement, no-member

    def test_get_clean_value(self):
        obj = BaseObjectFieldsTestClass()
        self.assertIsNone(obj.field1)
        obj.field1 = 21
        self.assertEqual(obj.field1, 21)

    def test_get_clean_value_error(self):
        with self.assertRaisesRegex(FieldValueError, 'field1: Error'):
            BaseObjectCleanValueError(field1='21')

    def test_del_attribute(self):
        obj = BaseObject(attr1=1)
        self.assertEqual(obj.attr1, 1)
        del obj.attr1
        with self.assertRaisesRegex(AttributeError, "'BaseObject' object has no attribute 'attr1'"):
            obj.attr1 # pylint: disable=pointless-statement

    def test_del_private_attribute(self):
        obj = BaseObject(_attr1=1)
        self.assertEqual(obj._attr1, 1)
        del obj._attr1
        with self.assertRaisesRegex(AttributeError, "'BaseObject' object has no attribute '_attr1'"):
            obj._attr1 # pylint: disable=pointless-statement

    def test_copy(self):
        obj1 = BaseObject(_private=1, public=2)
        obj2 = copy(obj1)
        self.assertNotEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)
        self.assertEqual(obj1.public, obj2.public)

    def test_copy_config(self):
        obj1 = BaseObject(config='Banana')
        obj2 = copy(obj1)
        self.assertEqual(obj1.config, 'Banana')
        with self.assertRaisesRegex(AttributeError, "'BaseObject' object has no attribute 'config'"):
            obj2.config # pylint: disable=pointless-statement

    def test_deepcopy(self):
        obj1 = BaseObject(_private=1, public=2)
        obj2 = deepcopy(obj1)
        self.assertNotEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)
        self.assertEqual(obj1.public, obj2.public)

    def test_deepcopy_config(self):
        obj1 = BaseObject(config='Banana')
        obj2 = deepcopy(obj1)
        self.assertEqual(obj1.config, 'Banana')
        with self.assertRaisesRegex(AttributeError, "'BaseObject' object has no attribute 'config'"):
            obj2.config # pylint: disable=pointless-statement

    def test_replace(self):
        obj1 = BaseObject(_private=1, public=2)
        obj2 = obj1.replace(_private=3, public=4)
        self.assertNotEqual(obj1, obj2)
        self.assertEqual(obj2._private, 3)
        self.assertEqual(obj2.public, 4)

    def test_replace_config(self):
        obj1 = BaseObject(config=1)
        obj2 = obj1.replace(public=4)
        with self.assertRaisesRegex(AttributeError, "'BaseObject' object has no attribute 'config'"):
            obj2.config # pylint: disable=pointless-statement


## Fake class for BaseDict tests.
class BaseDictTestClass(BaseDict):
    attr: int = None
    field = BaseField(attr_type=str)

class BaseDictFieldsTestClass(BaseDict):
    _private: str = 'Private'
    normal_attr = 'Normal'
    field1 = BaseField(attr_type=int)
    field_default = BaseField(attr_type=int, default=100)
    field_readonly = BaseField(attr_type=str, default='default_value', readonly=True)

class BaseDictInheritanceTestClass(BaseDictTestClass):
    attr_other: str = None

class BaseDictRequiredTestClass(BaseDict):
    required = BaseField(attr_type=int, required=True)

class BaseDictRequiredLazyTestClass(BaseDict):
    normal: bool = True
    required = BaseField(attr_type=int, required_lazy=True)

class BaseDictBoolTestClass(BaseDict):
    attr = BoolField()

class BaseDictUndefinedTestClass(BaseDict):
    attr = BaseField(default=Undefined, attr_type=str)

class FakeNPArray:
    def __eq__(self, obj: Any) -> bool:
        raise ValueError('The truth value of an array with more than one element is ambiguous. Use a.any() or a.all()')

class BaseDictNpArrayTestCase(BaseDict):
    attr: FakeNPArray = None

class NpArrayTestCase(TestCase):

    def test_equal_raise(self):
        obj = FakeNPArray()
        with self.assertRaisesRegex(ValueError, r'The truth value of an array with more than one element is ambiguous. Use a.any\(\) or a.all\(\)'):
            assert obj == ''

    def test_set_value(self):
        fake_np_array = FakeNPArray()

        obj = BaseDictNpArrayTestCase()
        obj.attr = fake_np_array
        self.assertIsNotNone(obj.attr)
        self.assertIsNotNone(obj['attr'])

        obj = BaseDictNpArrayTestCase(attr=fake_np_array)
        self.assertIsNotNone(obj.attr)
        self.assertIsNotNone(obj['attr'])

class BaseDictTestCase(TestCase):

    def test_init(self):
        obj = BaseDictTestClass()
        self.assertIsNone(obj.attr)
        self.assertIsNone(obj['field'])

    def test_attr_creation(self):
        obj = BaseDictTestClass()
        obj.attr = 1
        self.assertEqual(obj.attr, 1)
        self.assertEqual(obj['attr'], 1)
        obj['attr2'] = 10
        self.assertEqual(obj.attr2, 10) # pylint: disable=no-member
        self.assertEqual(obj['attr2'], 10)

    def test_fields(self):
        obj = BaseDictFieldsTestClass()
        self.assertEqual(obj._private, 'Private')
        self.assertEqual(obj.normal_attr, 'Normal')
        self.assertIsNone(obj.field1)
        self.assertIsNone(obj['field1'])
        self.assertEqual(obj.field_default, 100)
        self.assertEqual(obj['field_default'], 100)
        self.assertEqual(obj.field_readonly, 'default_value')
        self.assertEqual(obj['field_readonly'], 'default_value')
        self.assertRaises(FieldValueError, BaseDictFieldsTestClass, field_readonly='Changed')
        with self.assertRaisesRegex(FieldValueError, "The field 'field_readonly' value cannot be changed."):
            obj.field_readonly = 'Changed'
        with self.assertRaisesRegex(FieldValueError, "The field 'field_readonly' value cannot be changed."):
            obj['field_readonly'] = 'Changed'

    def test_private_attr(self):
        obj = BaseDictFieldsTestClass()
        self.assertNotIn('_private', obj)
        obj._other = 1
        self.assertNotIn('_other', obj)
        self.assertEqual(obj._other, 1)
        with self.assertRaisesRegex(KeyError, "Keys can't start with '_'"):
            obj['_other'] = 1

    def test_config_key(self):
        obj = BaseDict(config=1)
        with self.assertRaisesRegex(KeyError, 'config'):
            obj['config'] # pylint: disable=pointless-statement

        obj = BaseDict()
        with self.assertRaisesRegex(KeyError, 'The key cannot be called "config".'):
            obj['config'] = 1

    def test_update(self):
        obj = BaseDictTestClass()
        self.assertIsNone(obj.attr)
        self.assertIsNone(obj['field'])
        obj.update({'attr': 1, 'field': 'a', 'attr2': 2})
        self.assertEqual(obj.attr, 1)
        self.assertEqual(obj['attr'], 1)
        self.assertEqual(obj.attr2, 2) # pylint: disable=no-member
        self.assertEqual(obj['attr2'], 2)
        self.assertRaises(FieldValueError, obj.update, {'attr': '1', 'attr2': 2})

    def test_required(self):
        self.assertRaises(RequiredError, BaseDictRequiredTestClass)

    def test_required_lazy(self):
        obj = BaseDictRequiredLazyTestClass()
        self.assertRaisesRegex(
            RequiredError,
            'The required attribute is required.',
            obj.validate_required_fields
        )

    def test_default_receive_none(self):
        obj = BaseDictFieldsTestClass(field_default=None)
        self.assertEqual(obj.field_default, None)
        self.assertEqual(obj['field_default'], None)

    def test_value_persist_on_error(self):
        obj = BaseDictFieldsTestClass()
        try:
            obj['field_readonly'] = 'Changed'
        except FieldValueError:
            pass
        self.assertEqual(obj.field_readonly, 'default_value')
        self.assertEqual(obj['field_readonly'], 'default_value')

    def test_clean_value(self):
        base = BaseDictBoolTestClass(attr=1)
        # check clean value changed to bool True
        self.assertTrue(base.attr is True)
        self.assertTrue(base['attr'] is True)

        base = BaseDictBoolTestClass()
        base.attr = 1
        self.assertTrue(base.attr is True)
        self.assertTrue(base['attr'] is True)

        base = BaseDictBoolTestClass()
        base['attr'] = 1
        self.assertTrue(base.attr is True)
        self.assertTrue(base['attr'] is True)

    def test_del_attribute(self):
        obj = BaseDict(attr1=1)
        self.assertEqual(obj['attr1'], 1)
        self.assertEqual(obj.attr1, 1)
        del obj.attr1
        with self.assertRaisesRegex(KeyError, "'attr1'"):
            obj['attr1'] # pylint: disable=pointless-statement

        with self.assertRaisesRegex(AttributeError, "'BaseDict' object has no attribute 'attr1'"):
            obj.attr1 # pylint: disable=pointless-statement

    def test_del_private_attribute(self):
        obj = BaseDict(_attr1=1)
        # Private attributes don't turn into keys
        with self.assertRaisesRegex(KeyError, "'_attr1'"):
            obj['_attr1'] # pylint: disable=pointless-statement

        self.assertEqual(obj._attr1, 1)
        del obj._attr1
        with self.assertRaisesRegex(KeyError, "'_attr1'"):
            obj['_attr1'] # pylint: disable=pointless-statement

        with self.assertRaisesRegex(AttributeError, "'BaseDict' object has no attribute '_attr1'"):
            obj._attr1 # pylint: disable=pointless-statement

    def test_del_key(self):
        obj = BaseDict(attr1=1)
        self.assertEqual(obj['attr1'], 1)
        self.assertEqual(obj.attr1, 1)
        del obj['attr1']
        with self.assertRaisesRegex(KeyError, "'attr1'"):
            obj['attr1'] # pylint: disable=pointless-statement

        with self.assertRaisesRegex(AttributeError, "'BaseDict' object has no attribute 'attr1'"):
            obj.attr1 # pylint: disable=pointless-statement

    def test_fromkeys(self):
        obj = BaseDict(_private=1, public=2, excluded=3)
        dct = obj.fromkeys(['public', 'new'], 4)
        self.assertEqual(obj._private, dct._private)
        self.assertEqual(obj.public, dct.public)
        self.assertEqual(obj['public'], dct['public'])
        self.assertEqual(dct.new, 4)
        self.assertEqual(dct['new'], 4)

    def test_clear(self):
        obj = BaseDict(_private=1, public=2)
        self.assertDictEqual(obj, {'public': 2})
        self.assertEqual(obj._private, 1)
        self.assertEqual(obj.public, 2)
        obj.clear()
        self.assertDictEqual(obj, {})
        self.assertEqual(obj._private, 1)
        self.assertFalse(hasattr(obj, 'public'))

    def test_self_copy(self):
        obj1 = BaseDict(_private=1, public=2)
        obj2 = obj1.copy()
        self.assertDictEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)

    def test_copy(self):
        obj1 = BaseDict(_private=1, public=2)
        obj2 = copy(obj1)
        self.assertDictEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)

    def test_deepcopy(self):
        obj1 = BaseDict(_private=1, public=2)
        obj2 = deepcopy(obj1)
        self.assertDictEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)

    def test_pop(self):
        obj = BaseDict(_private=1, public=2)
        self.assertEqual(obj.pop('public'), 2)
        self.assertEqual(obj.pop('public', 4), 4)
        with self.assertRaisesRegex(KeyError, 'public'):
            obj.pop('public')

    def test_popitem(self):
        obj = BaseDict(_private=1, public=2)
        self.assertEqual(obj.popitem(), ('public', 2))
        with self.assertRaisesRegex(KeyError, r'popitem\(\)\: dictionary is empty'):
            obj.popitem()

    def test_replace(self):
        obj1 = BaseDict(_private=1, public=2)
        obj2 = obj1.replace(_private=3, public=4)
        self.assertNotEqual(obj1, obj2)
        self.assertEqual(obj2._private, 3)
        self.assertEqual(obj2.public, 4)
        self.assertEqual(obj2['public'], 4)

    def test_replace_config(self):
        obj1 = BaseDict()
        obj1.config = 1
        obj2 = obj1.replace(public=4)
        with self.assertRaisesRegex(AttributeError, "'BaseDict' object has no attribute 'config'"):
            obj2.config # pylint: disable=pointless-statement

    def test_init_with_undefined(self):
        obj = BaseDictUndefinedTestClass(attr='foo')
        self.assertEqual(obj.attr, 'foo')
        self.assertEqual(obj['attr'], 'foo')
        obj = BaseDictUndefinedTestClass(attr=Undefined.default_parse_string)
        self.assertEqual(obj.attr, Undefined)
        self.assertEqual(obj['attr'], Undefined)
        obj = BaseDictUndefinedTestClass(attr=Undefined)
        self.assertEqual(obj.attr, Undefined)
        self.assertEqual(obj['attr'], Undefined)
        obj = BaseDictUndefinedTestClass(attr=None)
        self.assertEqual(obj.attr, None)
        self.assertEqual(obj['attr'], None)

    def test_set_undefined(self):
        obj = BaseDictUndefinedTestClass()
        obj.attr = Undefined.default_parse_string
        self.assertEqual(obj.attr, Undefined)
        self.assertEqual(obj['attr'], Undefined)

        obj.attr = Undefined
        self.assertEqual(obj.attr, Undefined)
        self.assertEqual(obj['attr'], Undefined)


class BaseDictProperty(BaseDict):
    _to_date = BaseField(attr_type=str)

    @property
    def to_date(self):
        return self._to_date

    @to_date.setter
    def to_date(self, value):
        self._to_date = value

class BaseDictPropertyInherit(BaseDictProperty):
    pass


class BaseDictPropertyTestCase(TestCase):

    def test_property_setattr_init(self):
        obj = BaseDictProperty(to_date='Test')
        self.assertEqual(obj.to_date, 'Test')
        self.assertEqual(obj._to_date, 'Test')
        self.assertNotIn('to_date', obj)

    def test_property_setattr(self):
        obj = BaseDictProperty()
        obj.to_date = 'Test'
        self.assertEqual(obj.to_date, 'Test')
        self.assertEqual(obj._to_date, 'Test')
        self.assertNotIn('to_date', obj)

    def test_property_setitem(self):
        obj = BaseDictProperty()
        obj['to_date'] = 'Test'
        self.assertEqual(obj.to_date, 'Test')
        self.assertEqual(obj._to_date, 'Test')
        self.assertNotIn('to_date', obj)


class BaseObjectConfig(BaseObject):
    class Config:
        value: int = 1

class BaseDictConfig(BaseDict):
    class Config:
        value: int = 1

    config: Config


class MetaClassConfigTestCase(TestCase):

    def setUp(self) -> None:
        # pylint: disable=no-member
        BaseObjectConfig.config.value = 1
        BaseDictConfig.config.value = 1

    def test_object_config_attribute(self):
        # pylint: disable=no-member
        self.assertEqual(BaseObjectConfig.config.value, 1)
        obj = BaseObjectConfig()
        self.assertEqual(obj.config.value, BaseObjectConfig.config.value)

    def test_object_config_singleton(self):
        # pylint: disable=no-member
        obj = BaseObjectConfig()
        obj.config.value = 3
        self.assertEqual(BaseObjectConfig.config.value, 3)
        self.assertEqual(obj.config.value, BaseObjectConfig.config.value)

    def test_object_config_delete(self):
        obj = BaseObjectConfig()
        with self.assertRaisesRegex(AttributeError, "type object 'BaseObjectConfig' has no attribute 'Config'"):
            BaseObjectConfig.Config # pylint: disable=pointless-statement

        with self.assertRaisesRegex(AttributeError, "'BaseObjectConfig' object has no attribute 'Config'"):
            obj.Config # pylint: disable=pointless-statement

    def test_dict_config_attribute(self):
        self.assertEqual(BaseDictConfig.config.value, 1)
        obj = BaseDictConfig()
        self.assertEqual(obj.config.value, BaseDictConfig.config.value)

    def test_dict_config_singleton(self):
        obj = BaseDictConfig()
        obj.config.value = 3
        self.assertEqual(BaseDictConfig.config.value, 3)
        self.assertEqual(obj.config.value, BaseDictConfig.config.value)

    def test_dict_config_delete(self):
        obj = BaseDictConfig()
        with self.assertRaisesRegex(AttributeError, "type object 'BaseDictConfig' has no attribute 'Config'"):
            BaseDictConfig.Config # pylint: disable=pointless-statement

        with self.assertRaisesRegex(AttributeError, "'BaseDictConfig' object has no attribute 'Config'"):
            obj.Config # pylint: disable=pointless-statement

    def test_dict_config_no_key(self):
        obj = BaseDictConfig()
        with self.assertRaisesRegex(KeyError, 'config'):
            obj['config'] # pylint: disable=pointless-statement

    def test_dict_config_attribute_copy(self):
        obj1 = BaseDictConfig()
        obj1.config = copy(obj1.config)
        obj1.config.value = 2
        obj2 = BaseDictConfig()
        obj3 = BaseDictConfig(**obj1)
        self.assertNotEqual(obj1.config, obj2.config)
        self.assertNotEqual(obj1.config, obj3.config)
        self.assertEqual(obj1.config.value, 2)
        self.assertEqual(obj2.config.value, 1)
        self.assertEqual(obj3.config.value, 1)

    def test_return_self(self):
        self.assertDictEqual(
            MetaClass.__call__.__annotations__,
            {'args': tuple, 'kwargs': dict, 'return': Self}
        )

        self.assertDictEqual(
            MetaClass.__new__.__annotations__,
            {'name': str, 'bases': tuple, 'attrs': dict, 'return': Self}
        )


class FrozenObject(BaseObject):
    class Config:
        frozen: bool = True


class FrozenObjectTestCase(TestCase):

    def setUp(self) -> None:
        self.message = 'Class everysk.core._tests.object.FrozenObject is frozen and cannot be modified.'

    def test_create(self):
        obj = FrozenObject(attr=1, attr2='Banana')
        self.assertEqual(obj.attr, 1)
        self.assertEqual(obj.attr2, 'Banana')

    def test_delete(self):
        obj = FrozenObject(attr=1)
        with self.assertRaisesRegex(AttributeError, self.message):
            del obj.attr

    def test_create_new_attribute(self):
        obj = FrozenObject(attr=1)
        with self.assertRaisesRegex(AttributeError, self.message):
            obj.attr2 = 2

    def test_create_update_attribute(self):
        obj = FrozenObject(attr=1)
        with self.assertRaisesRegex(AttributeError, self.message):
            obj.attr = 2

    def test_copy(self):
        obj1 = FrozenObject(_private=1, public=2)
        obj2 = copy(obj1)
        self.assertNotEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)
        self.assertEqual(obj1.public, obj2.public)
        self.assertTrue(obj1._is_frozen)
        self.assertTrue(obj2._is_frozen)

    def test_deepcopy(self):
        obj1 = FrozenObject(_private=1, public=2)
        obj2 = deepcopy(obj1)
        self.assertNotEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)
        self.assertEqual(obj1.public, obj2.public)
        self.assertTrue(obj1._is_frozen)
        self.assertTrue(obj2._is_frozen)

    def test_replace(self):
        obj1 = FrozenObject(_private=1, public=2)
        obj2 = obj1.replace(_private=3, public=4)
        self.assertNotEqual(obj1, obj2)
        self.assertEqual(obj2._private, 3)
        self.assertEqual(obj2.public, 4)
        self.assertTrue(obj2._is_frozen)


class FrozenDict(BaseDict):
    class Config:
        frozen: bool = True


class FrozenDictTestCase(TestCase):
    def setUp(self) -> None:
        self.message = 'Class everysk.core._tests.object.FrozenDict is frozen and cannot be modified.'

    def test_create(self):
        obj = FrozenDict(attr=1, attr2='Banana')
        self.assertEqual(obj.attr, 1)
        self.assertEqual(obj.attr2, 'Banana')
        self.assertEqual(obj['attr'], 1)
        self.assertEqual(obj['attr2'], 'Banana')

    def test_delete(self):
        obj = FrozenDict(attr=1, attr2='Banana')
        with self.assertRaisesRegex(AttributeError, self.message):
            del obj.attr

        with self.assertRaisesRegex(AttributeError, self.message):
            del obj['attr2']

    def test_create_new_attribute(self):
        obj = FrozenDict()
        with self.assertRaisesRegex(AttributeError, self.message):
            obj.attr = 2

        with self.assertRaisesRegex(AttributeError, self.message):
            obj['attr'] = 2

    def test_create_update_attribute(self):
        obj = FrozenDict(attr=1, attr2='Banana')
        with self.assertRaisesRegex(AttributeError, self.message):
            obj.attr = 2

        with self.assertRaisesRegex(AttributeError, self.message):
            obj['attr'] = 2

    def test_clear(self):
        obj = FrozenDict(attr=1, attr2='Banana')
        with self.assertRaisesRegex(AttributeError, self.message):
            obj.clear()

    def test_self_copy(self):
        obj1 = FrozenDict(_private=1, public=2)
        obj2 = obj1.copy()
        self.assertDictEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)
        self.assertTrue(obj2.__dict__['_is_frozen'])
        self.assertTrue(obj1._is_frozen)
        self.assertTrue(obj2._is_frozen)

    def test_copy(self):
        obj1 = FrozenDict(_private=1, public=2)
        obj2 = copy(obj1)
        self.assertDictEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)
        self.assertTrue(obj2.__dict__['_is_frozen'])
        self.assertTrue(obj1._is_frozen)
        self.assertTrue(obj2._is_frozen)

    def test_deepcopy(self):
        obj1 = FrozenDict(_private=1, public=2)
        obj2 = deepcopy(obj1)
        self.assertDictEqual(obj1, obj2)
        self.assertEqual(obj1._private, obj2._private)
        self.assertTrue(obj2.__dict__['_is_frozen'])
        self.assertTrue(obj1._is_frozen)
        self.assertTrue(obj2._is_frozen)

    def test_pop(self):
        obj = FrozenDict(_private=1, public=2)
        with self.assertRaisesRegex(AttributeError, self.message):
            obj.pop('public')

    def test_popitem(self):
        obj = FrozenDict(_private=1, public=2)
        with self.assertRaisesRegex(AttributeError, self.message):
            obj.popitem()

    def test_update(self):
        obj = FrozenDict(_private=1, public=2)
        with self.assertRaisesRegex(AttributeError, self.message):
            obj.update({'attr': 1})

    def test_replace(self):
        obj1 = FrozenDict(_private=1, public=2)
        obj2 = obj1.replace(_private=3, public=4)
        self.assertNotEqual(obj1, obj2)
        self.assertEqual(obj2._private, 3)
        self.assertEqual(obj2.public, 4)
        self.assertEqual(obj2['public'], 4)
        self.assertTrue(obj2._is_frozen)


class FakeBaseObject(BaseObject):
    class Config(BaseObject):
        f1: str = 'FakeBaseObject'
        f2 =  BaseField(attr_type=str, default='FakeBaseObject')

class FakeBaseDict(BaseDict):
    class Config(BaseObject):
        f1: str = 'FakeBaseDict'
        f2 = BaseField(attr_type=str, default='FakeBaseDict')

class Fake01PythonClass:
    class Config:
        pass

class Fake02PythonClass:
    class Config:
        pass

class ConfigHashTestCase(TestCase):
    # https://everysk.atlassian.net/browse/COD-2746

    def test_different_values(self):
        # pylint: disable=no-member
        self.assertEqual(FakeBaseObject.config.f1, 'FakeBaseObject')
        self.assertEqual(FakeBaseObject.config.f2, 'FakeBaseObject')
        self.assertEqual(FakeBaseDict.config.f1, 'FakeBaseDict')
        self.assertEqual(FakeBaseDict.config.f2, 'FakeBaseDict')

    def test_python_qualname(self):
        self.assertEqual(Fake01PythonClass.Config.__name__, 'Config')
        self.assertEqual(Fake01PythonClass.Config.__qualname__, 'Fake01PythonClass.Config')
        self.assertEqual(Fake02PythonClass.Config.__name__, 'Config')
        self.assertEqual(Fake02PythonClass.Config.__qualname__, 'Fake02PythonClass.Config')


class MetaClassParent(BaseObject):
    p01 = BaseField(attr_type=str, default='Parent field')
    p02: str = '1'
    p03 = 1
    p04: float

    @property
    def prop(self):
        return self.p01

    @prop.setter
    def prop(self, value: str) -> None:
        self.p01 = value

    def func(self):
        pass


class MetaClassChild(MetaClassParent):
    p01 = BaseField(attr_type=str, default='Child field')
    c02: str = '1'
    c03 = 1
    c04: float


class MetaClassAttributesTestCase(TestCase):

    def test_class_attributes(self):
        attrs = getattr(MetaClassParent, MetaClass._attr_name)
        self.assertDictEqual(
            attrs,
            {
                '_errors': dict,
                '_is_frozen': bool,
                '_silent': bool,
                'p01': BaseField(attr_type=str, default='Parent field'),
                'p02': str,
                'p03': int,
                'p04': float
            }
        )

    def test_class_annotations(self):
        self.assertDictEqual(
            MetaClassParent.__annotations__,
            {
                'p01': str,
                'p02': str,
                'p03': int,
                'p04': float
            }
        )

    def test_class_inheritance_attributes(self):
        attrs = getattr(MetaClassChild, MetaClass._attr_name)
        self.assertDictEqual(
            attrs,
            {
                '_errors': dict,
                '_is_frozen': bool,
                '_silent': bool,
                'p01': BaseField(attr_type=str, default='Child field'),
                'p02': str,
                'p03': int,
                'p04': float,
                'c02': str,
                'c03': int,
                'c04': float
            }
        )

    def test_class_inheritance_annotations(self):
        self.assertDictEqual(
            MetaClassChild.__annotations__,
            {
                'p01': str,
                'c02': str,
                'c03': int,
                'c04': float
            }
        )

    def test_class_inheritance_default_value(self):
        self.assertEqual(MetaClassChild.p01, 'Child field')
        self.assertEqual(MetaClassChild.p02, '1')
        self.assertEqual(MetaClassChild.p03, 1)
        self.assertIsNone(MetaClassChild.p04)
        self.assertIsNone(MetaClassChild().func())

    def test_class_inheritance_changed_value(self):
        old_parent_p02 = MetaClassParent.p02
        old_child_p02 = MetaClassChild.p02
        MetaClassParent.p02 = 'First'
        self.assertEqual(MetaClassParent.p02, 'First')
        self.assertEqual(MetaClassChild.p02, 'First')
        MetaClassChild.p02 = 'Second'
        self.assertEqual(MetaClassParent.p02, 'First')
        self.assertEqual(MetaClassChild.p02, 'Second')
        MetaClassParent.p02 = old_parent_p02
        MetaClassChild.p02 = old_child_p02

    def test_obj_attributes(self):
        obj = MetaClassParent()
        attrs = getattr(obj, MetaClass._attr_name)
        self.assertDictEqual(
            attrs,
            {
                '_errors': dict,
                '_is_frozen': bool,
                '_silent': bool,
                'p01': BaseField(attr_type=str, default='Parent field'),
                'p02': str,
                'p03': int,
                'p04': float
            }
        )

    def test_obj_annotations(self):
        obj = MetaClassParent()
        self.assertDictEqual(
            obj.__annotations__,
            {
                'p01': str,
                'p02': str,
                'p03': int,
                'p04': float
            }
        )

    def test_obj_inheritance_attributes(self):
        obj = MetaClassChild()
        attrs = getattr(obj, MetaClass._attr_name)
        self.assertDictEqual(
            attrs,
            {
                '_errors': dict,
                '_is_frozen': bool,
                '_silent': bool,
                'p01': BaseField(attr_type=str, default='Child field'),
                'p02': str,
                'p03': int,
                'p04': float,
                'c02': str,
                'c03': int,
                'c04': float
            }
        )

    def test_obj_inheritance_annotations(self):
        obj = MetaClassChild()
        self.assertDictEqual(
            obj.__annotations__,
            {
                'p01': str,
                'c02': str,
                'c03': int,
                'c04': float
            }
        )

    def test_obj_inheritance_default_value(self):
        obj = MetaClassChild()
        self.assertEqual(obj.p01, 'Child field')
        self.assertEqual(obj.p02, '1')
        self.assertEqual(obj.p03, 1)
        self.assertIsNone(obj.p04)

    def test_obj_inheritance_changed_value(self):
        parent = MetaClassParent()
        child = MetaClassChild()
        parent.p02 = 'First'
        self.assertEqual(parent.p02, 'First')
        self.assertEqual(child.p02, '1')
        child.p02 = 'Second'
        self.assertEqual(parent.p02, 'First')
        self.assertEqual(child.p02, 'Second')

    def test_property_value(self):
        # https://everysk.atlassian.net/browse/COD-3457
        obj = MetaClassParent(p01='01')
        self.assertEqual(obj.p01, '01')
        self.assertEqual(obj.prop, '01')

        obj = MetaClassParent(prop='02')
        self.assertEqual(obj.p01, '02')
        self.assertEqual(obj.prop, '02')


class BeforeInitTestCase(TestCase):

    class BeforeInitBaseObject(BaseObject):
        key: str = None

        @classmethod
        def __before_init__(cls, **kwargs: dict) -> dict:
            return {'key': 'before'}

    class BeforeInitBaseDict(BaseDict):
        key: str = None

        @classmethod
        def __before_init__(cls, **kwargs: dict) -> dict:
            return {'key': 'before'}

    class BeforeInitBaseObjectNone(BaseObject):
        key: str = None

        @classmethod
        def __before_init__(cls, **kwargs: dict) -> dict:
            return None

    class BeforeInitBaseDictNone(BaseDict):
        key: str = None

        @classmethod
        def __before_init__(cls, **kwargs: dict) -> dict:
            return None

    def test_base_object(self):
        obj = BeforeInitTestCase.BeforeInitBaseObject(key='test')
        self.assertEqual(obj.key, 'before')

    def test_base_object_none(self):
        obj = BeforeInitTestCase.BeforeInitBaseObjectNone(key='test')
        self.assertEqual(obj.key, 'test')

    def test_base_dict(self):
        obj = BeforeInitTestCase.BeforeInitBaseDict(key='test')
        self.assertEqual(obj.key, 'before')
        self.assertEqual(obj['key'], 'before')

    def test_base_dict_none(self):
        obj = BeforeInitTestCase.BeforeInitBaseDictNone(key='test')
        self.assertEqual(obj.key, 'test')
        self.assertEqual(obj['key'], 'test')


class AfterInitTestCase(TestCase):

    class AfterInitBaseObject(BaseObject):
        key: str = None
        def __after_init__(self) -> None:
            self.key = 'after'

    class AfterInitBaseDict(BaseDict):
        key: str = None
        def __after_init__(self) -> None:
            self.key = 'after'

    def test_after_base_object(self):
        obj = BaseObject(key='test')
        self.assertEqual(obj.key, 'test')

        obj = AfterInitTestCase.AfterInitBaseObject(key='test')
        self.assertEqual(obj.key, 'after')

    def test_after_base_dict(self):
        obj = BaseDict(key='test')
        self.assertEqual(obj.key, 'test')
        self.assertEqual(obj['key'], 'test')

        obj = AfterInitTestCase.AfterInitBaseDict(key='test')
        self.assertEqual(obj.key, 'after')
        self.assertEqual(obj['key'], 'after')


class SilentTestCase(TestCase):

    class SilentBaseObject(BaseObject):
        f1: bool = False

        @classmethod
        def __before_init__(cls, **kwargs: dict) -> dict:
            if 'before' in kwargs:
                raise ValueError('before')
            return kwargs

        def __after_init__(self) -> None:
            if hasattr(self, 'after'):
                raise ValueError('after')

        def _init_error_handler(self, kwargs: dict, errors: dict[str, Exception]) -> None:
            for key, error in errors.items():
                if error:
                    self.error_found = f'{key}: {error}'

    def test_init_error(self):
        self.assertRaisesRegex(FieldValueError, "Key f1 must be <class 'bool'>.", SilentTestCase.SilentBaseObject, f1=2)

    def test_init_error_silent(self):
        obj = SilentTestCase.SilentBaseObject(f1=2, silent=True)
        self.assertEqual(obj.error_found, "init: Key f1 must be <class 'bool'>.")

    def test_before_init_error(self):
        self.assertRaisesRegex(ValueError, "before", SilentTestCase.SilentBaseObject, before=2)

    def test_before_init_error_silent(self):
        obj = SilentTestCase.SilentBaseObject(before=2, silent=True)
        self.assertEqual(obj.error_found, "before_init: before")

    def test_after_init_error(self):
        self.assertRaisesRegex(ValueError, "after", SilentTestCase.SilentBaseObject, after=2)

    def test_after_init_error_silent(self):
        obj = SilentTestCase.SilentBaseObject(after=2, silent=True)
        self.assertEqual(obj.error_found, "after_init: after")



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# pylint: disable=protected-access
from concurrent.futures import ThreadPoolExecutor
from os import environ
from time import sleep
from uuid import uuid4
from unittest import TestCase, mock
from redis import exceptions, client
from everysk.core import redis as redis_module
from everysk.core.redis import (
    RedisCache, RedisCacheCompressed, RedisChannel, RedisClient,
    RedisEmptyListError, RedisList, RedisLock
)


class RedisClientTestCase(TestCase):

    def setUp(self) -> None:
        RedisClient._connection = None

    @mock.patch.object(redis_module, 'log')
    def test_environ(self, log: mock.MagicMock):
        environ['REDIS_HOST'] = '0.1.0.1'
        environ['REDIS_PORT'] = '1234'
        redis = RedisClient()
        self.assertEqual(redis.host, environ['REDIS_HOST'])
        self.assertEqual(redis.port, int(environ['REDIS_PORT']))
        del environ['REDIS_HOST']
        del environ['REDIS_PORT']
        log.debug.assert_called_once_with('Connecting on Redis(%s).....', redis.host)

    @mock.patch.object(redis_module, 'log')
    def test_connection_property(self, log: mock.MagicMock):
        redis = RedisClient()
        self.assertEqual(redis.connection, redis._connection)
        self.assertEqual(redis.connection, RedisClient._connection)
        log.debug.assert_called_once()

    @mock.patch.object(redis_module, 'log')
    def test_hash_key(self, log: mock.MagicMock):
        redis = RedisClient()
        self.assertEqual(
            redis.get_hash_key('a'),
            'ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb'
        )
        # Check again to guarantee the same hash
        self.assertEqual(
            redis.get_hash_key('a'),
            'ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb'
        )
        log.debug.assert_called_once()

    @mock.patch.object(redis_module, 'log')
    def test_connect(self, log: mock.MagicMock):
        redis = RedisClient()
        log.debug.assert_called_once()
        RedisClient._connection = None
        self.assertIsNone(RedisClient._connection)
        redis._connect()
        self.assertIsNotNone(RedisClient._connection)

    @mock.patch.object(RedisClient, '_connect')
    def test_connection_error_connection(self, _connect: mock.MagicMock):
        with mock.patch.object(RedisClient, '_connection', spec=redis_module.Redis) as _connection:
            _connection.ping.side_effect = exceptions.ConnectionError
            connection = RedisClient().connection # pylint: disable=unused-variable
            _connection.ping.assert_called_once_with()
        _connect.assert_called_once_with()

    @mock.patch.object(redis_module.log, 'debug', mock.MagicMock)
    def test_flush_all(self):
        cli = RedisClient()
        # Generate a random key to test
        key = uuid4().hex
        cli.connection.set(key, 'Flush-test')
        cli.flush_all()
        self.assertIsNone(cli.connection.get(key))


@mock.patch.object(redis_module, 'log', mock.MagicMock())
class RedisCacheTestCase(TestCase):

    def setUp(self) -> None:
        RedisClient._connection = None

    def test_get_hash_key(self):
        obj = RedisCache()
        hash_key = '2c70e12b7a0646f92279f427c7b38e7334d8e5389cff167a1dc30e73f826b683'
        self.assertEqual(obj.get_hash_key('key'), hash_key)
        # Must always be the same value
        self.assertEqual(obj.get_hash_key('key'), hash_key)

    def test_get_set_delete(self):
        obj = RedisCache()
        obj.delete('key')
        self.assertIsNone(obj.get('key'))
        obj.set('key', 'My value.')
        self.assertEqual(obj.get('key'), b'My value.')
        obj.set('key', 1)
        self.assertEqual(obj.get('key'), b'1')
        obj.set('key', b'1.1')
        self.assertEqual(obj.get('key'), b'1.1')
        obj.delete('key')
        self.assertIsNone(obj.get('key'))

    def test_set_error(self):
        obj = RedisCache()
        self.assertRaises(ValueError, obj.set, 'key', ())
        self.assertRaises(ValueError, obj.set, 'key', [])
        self.assertRaises(ValueError, obj.set, 'key', {})
        self.assertRaises(ValueError, obj.set, 'key', obj)


class RedisCacheGetSetTestCase(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        RedisClient._connection = None
        with mock.patch.object(redis_module.log, 'debug') as debug:
            cls.cache = RedisCache()
            debug.assert_called_once_with('Connecting on Redis(%s).....', '0.0.0.0')

        cls.key = 'redis-get-set-test-case'

    def setUp(self) -> None:
        self.cache.delete(self.key)
        self.mock = mock.MagicMock()

    def func(self, **kwargs):
        sleep(0.1)
        self.mock(**kwargs)
        return f'return: {kwargs}'

    def test_redis_get_set(self):
        obj01 = RedisCache()
        obj02 = RedisCache()
        with ThreadPoolExecutor() as executor:
            task01 = executor.submit(obj01.get_set, key=self.key, func=self.func, keyword='value01')
            sleep(0.05) # This is just to guarantee that value1 will always be called first
            task02 = executor.submit(obj02.get_set, key=self.key, func=self.func, keyword='value02')
            self.assertEqual(task01.result(), task02.result())
        self.mock.assert_called_once_with(keyword='value01')

    @mock.patch.object(redis_module.log, 'error')
    def test_get_set_func_error_both_exit(self, error: mock.MagicMock):
        self.mock.side_effect = AttributeError
        obj01 = RedisCache()
        obj02 = RedisCache()
        with ThreadPoolExecutor() as executor:
            task01 = executor.submit(obj01.get_set, key=self.key, func=self.func, keyword='value01')
            sleep(0.05) # This is just to guarantee that value1 will always be called first
            task02 = executor.submit(obj02.get_set, key=self.key, func=self.func, keyword='value02')
            self.assertIsNone(task01.result())
            self.assertIsNone(task02.result())

        self.mock.assert_called_once_with(keyword='value01')
        error.assert_called_once()

    @mock.patch.object(redis_module.log, 'error')
    def test_get_set_func_error(self, error: mock.MagicMock):
        self.mock.side_effect = AttributeError
        result = RedisCache().get_set(key=self.key, func=self.func)
        self.assertIsNone(result)
        self.mock.assert_called_once_with()
        error.assert_called_once()


@mock.patch.object(redis_module, 'log', mock.MagicMock())
class RedisCacheCompressedTestCase(TestCase):

    def setUp(self) -> None:
        RedisClient._connection = None

    def test_get_hash_key(self):
        obj = RedisCacheCompressed()
        hash_key = '2c70e12b7a0646f92279f427c7b38e7334d8e5389cff167a1dc30e73f826b683'
        self.assertEqual(obj.get_hash_key('key'), hash_key)
        # Must always be the same value
        self.assertEqual(obj.get_hash_key('key'), hash_key)

    def test_get_set_delete(self):
        obj = RedisCacheCompressed()
        self.assertIsNone(obj.get('key'))
        lst = [1, 'test', 1.1, {'key': 'value'}]
        obj.set('key', lst)
        self.assertListEqual(obj.get('key'), lst)
        dct = {'lst': [1, 't'], 'dct': {'key': 'value'}, 'key': 1}
        obj.set('key', dct)
        self.assertDictEqual(obj.get('key'), dct)
        obj.delete('key')
        self.assertIsNone(obj.get('key'))


@mock.patch.object(redis_module, 'log', mock.MagicMock())
class RedisListTestCase(TestCase):

    def setUp(self) -> None:
        RedisClient._connection = None
        self.name = 'redis-list-test-case'

    def test_push_pop(self):
        lst = RedisList(name=self.name)
        lst.clear() # Clear all elements
        self.assertRaises(RedisEmptyListError, lst.pop)
        obj = {'lst': [1, 1.1, 'test', True]}
        lst.push(obj)
        self.assertEqual(lst.pop(), obj)

    def test_bpop_timeout(self):
        lst = RedisList(name=self.name)
        lst.clear() # Clear all elements
        self.assertRaises(RedisEmptyListError, lst.bpop, 0.5)

    def test_bpop(self):
        lst = RedisList(name=self.name)
        lst.clear() # Clear all elements
        obj = {'lst': [1, 1.1, 'test', True]}
        lst.push(obj)
        self.assertTupleEqual(lst.bpop(0.5), (self.name, obj))

    def test_clear(self):
        lst = RedisList(name=self.name)
        lst.clear() # Clear all elements
        self.assertRaises(RedisEmptyListError, lst.pop)
        obj = {'lst': [1, 1.1, 'test', True]}
        lst.push(obj)
        lst.clear()
        self.assertRaises(RedisEmptyListError, lst.pop)


@mock.patch.object(redis_module, 'log', mock.MagicMock())
@mock.patch.object(RedisChannel, '_channel', spec=client.PubSub)
class RedisChannelTestCase(TestCase):

    def setUp(self) -> None:
        self.name = 'tests'
        RedisClient._connection = None

    @mock.patch.object(RedisClient, '_connection', spec=redis_module.Redis)
    def test_send(self, _connection: mock.MagicMock, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        channel.send('Hi')
        _connection.publish.assert_called_once_with(channel.name, 'Hi')
        _channel.assert_not_called()

    @mock.patch.object(RedisClient, '_connection', spec=redis_module.Redis)
    def test_channel(self, _connection: mock.MagicMock, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        channel._channel = None
        _connection.pubsub.return_value = _channel
        channel.channel # Call the property - pylint: disable=pointless-statement
        _connection.pubsub.assert_called_once_with()
        _channel.subscribe.assert_called_once_with(channel.name)

    def test_parse_message(self, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        message = {'channel': self.name.encode('utf-8'), 'data': b'data'}
        self.assertTupleEqual(channel.parse_message(message), (self.name, 'data'))
        _channel.assert_not_called()

    def test_consume_process_message(self, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        message = {'channel': self.name.encode('utf-8'), 'data': b'data'}
        _channel.listen.return_value = [message]
        channel.process_message = mock.MagicMock()
        channel.consume()
        channel.process_message.assert_called_once_with('data')

    def test_process(self, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        self.assertIsNone(channel.process_message('data'))
        _channel.assert_not_called()

    def test_consume_callback(self, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        message = {'channel': self.name.encode('utf-8'), 'data': b'data'}
        _channel.listen.return_value = [message]
        callback = mock.MagicMock()
        channel.consume(callback)
        callback.assert_called_once_with('data')

    def test_consume_break(self, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        message = {'channel': self.name.encode('utf-8'), 'data': b'data'}
        exit_message = {'channel': self.name.encode('utf-8'), 'data': RedisChannel.exit_message}
        _channel.listen.return_value = [exit_message, message]
        callback = mock.MagicMock()
        channel.consume(callback)
        callback.assert_not_called()


@mock.patch.object(redis_module, 'log', mock.MagicMock())
class RedisLockTestCase(TestCase):

    def setUp(self) -> None:
        self.name = 'tests'
        RedisClient._connection = None

    def tearDown(self) -> None:
        RedisLock(name=self.name).release(force=True)

    def test_acquire(self):
        lock01 = RedisLock(name=self.name)
        lock02 = RedisLock(name=self.name)
        self.assertTrue(lock01.acquire())
        self.assertFalse(lock02.acquire())

    @mock.patch.object(RedisLock, '_lock')
    def test_acquire_blocking(self, lock: mock.MagicMock):
        lock01 = RedisLock(name=self.name)
        lock01.acquire(blocking=True, blocking_timeout=1.0)
        lock.acquire.assert_called_once_with(blocking=True, blocking_timeout=1.0)

    def test_timeout(self):
        lock01 = RedisLock(name=self.name, timeout=1)
        lock02 = RedisLock(name=self.name, timeout=1)
        self.assertTrue(lock01.acquire())
        self.assertFalse(lock02.acquire())
        sleep(1.5) # With 1 second the test breaks sometimes
        self.assertTrue(lock02.acquire())

    def test_owned(self):
        lock01 = RedisLock(name=self.name)
        lock02 = RedisLock(name=self.name)
        lock01.acquire()
        lock02.acquire()
        self.assertTrue(lock01.owned())
        self.assertFalse(lock02.owned())

    def test_release(self):
        lock01 = RedisLock(name=self.name)
        lock02 = RedisLock(name=self.name)
        lock01.acquire()
        lock02.acquire()
        self.assertFalse(lock02.release())
        self.assertFalse(lock02.acquire())
        self.assertTrue(lock01.release())
        self.assertTrue(lock02.acquire())

    def test_release_force(self):
        lock01 = RedisLock(name=self.name)
        lock02 = RedisLock(name=self.name)
        lock01.acquire()
        lock02.acquire()
        self.assertTrue(lock02.release(force=True))
        self.assertTrue(lock02.acquire())
        self.assertFalse(lock01.release())
        self.assertFalse(lock01.acquire())



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###########################################
# API RESOURCE
###########################################
from everysk.api.api_resources._tests.api_resource import APIResourceTestCase as EveryskLibAPIResourceTestCase # pylint: disable=unused-import

###########################################
# FILE
###########################################
from everysk.api.api_resources._tests.file import APIFileTestCase as EveryskLibAPIFileTestCase # pylint: disable=unused-import

###########################################
# CUSTOM INDEX
###########################################
from everysk.api.api_resources._tests.custom_index import APICustomIndexTestCase as EveryskLibAPICustomIndexTestCase # pylint: disable=unused-import

###########################################
# PRIVATE SECURITY
###########################################
from everysk.api.api_resources._tests.private_security import APIPrivateSecurityTestCase as EveryskLibAPIPrivateSecurityTestCase # pylint: disable=unused-import

###########################################
# WORKSPACE
###########################################
from everysk.api.api_resources._tests.workspace import APIWorkspaceTestCase as EveryskLibAPIWorkspaceTestCase # pylint: disable=unused-import

###########################################
# REPORT TEMPLATE
###########################################
from everysk.api.api_resources._tests.report_template import APIReportTemplateTestCase as EveryskLibAPIReportTemplateTestCase # pylint: disable=unused-import

###########################################
# PORTFOLIO
###########################################
from everysk.api.api_resources._tests.portfolio import APIPortfolioTestCase as EveryskLibAPIPortfolioTestCase # pylint: disable=unused-import

###########################################
# DATASTORE
###########################################
from everysk.api.api_resources._tests.datastore import APIDatastoreTestCase as EveryskLibAPIDatastoreTestCase # pylint: disable=unused-import

###########################################
# WORKER EXECUTION
###########################################
from everysk.api.api_resources._tests.worker_execution import APIWorkerExecutionTestCase as EveryskLibAPIWorkerExecutionTestCase # pylint: disable=unused-import

###########################################
# PARSER
###########################################
from everysk.api.api_resources._tests.parser import APIParserTestCase as EveryskLibAPIParserTestCase # pylint: disable=unused-import

###########################################
# WORKFLOW EXECUTION
###########################################
from everysk.api.api_resources._tests.workflow_execution import APIWorkflowExecutionTestCase as EveryskLibAPIWorkflowExecutionTestCase # pylint: disable=unused-import

###########################################
# WORKFLOW
###########################################
from everysk.api.api_resources._tests.workflow import APIWorkflowTestCase as EveryskLibAPIWorkflowTestCase # pylint: disable=unused-import

###########################################
# CALCULATION
###########################################
from everysk.api.api_resources._tests.calculation import APICalculationTestCase as EveryskLibAPICalculationTestCase # pylint: disable=unused-import

###########################################
# MARKET DATA
###########################################
from everysk.api.api_resources._tests.market_data import APIMarketDataTestCase as EveryskLibAPIMarketDataTestCase # pylint: disable=unused-import

###########################################
# REPORT
###########################################
from everysk.api.api_resources._tests.report import APIReportTestCase as EveryskLibAPIReportTestCase # pylint: disable=unused-import

###########################################
# RETRIEVABLE API
###########################################
from everysk.api.api_resources._tests.api_resource import RetrievableAPIResourceTestCase as EveryskLibRetrievableAPIResourceTestCase # pylint: disable=unused-import

###########################################
# LISTABLE API
###########################################
from everysk.api.api_resources._tests.api_resource import ListableAPIResourceTestCase as EveryskLibListableAPIResourceTestCase # pylint: disable=unused-import

###########################################
# CREATABLE API
###########################################
from everysk.api.api_resources._tests.api_resource import CreatableAPIResourceTestCase as EveryskLibCreatableAPIResourceTestCase  # pylint: disable=unused-import

###########################################
# FILTERABLE API
###########################################
from everysk.api.api_resources._tests.api_resource import FilterableAPIResourceTestCase as EveryskLibFilterableAPIResourceTestCase # pylint: disable=unused-import

###########################################
# UPDATABLE API
###########################################
from everysk.api.api_resources._tests.api_resource import UpdatableAPIResourceTestCase as EveryskLibUpdatableAPIResourceTestCase # pylint: disable=unused-import

###########################################
# DELETABLE API
###########################################
from everysk.api.api_resources._tests.api_resource import DeletableAPIResourceTestCase as EveryskLibDeletableAPIResourceTestCase # pylint: disable=unused-import



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
import unittest
from everysk.config import settings

from everysk.core.datetime import DateTime, Date

from everysk.core.exceptions import RequiredError, FieldValueError

from everysk.sdk.entities.datastore.base import Datastore

from everysk.sdk.base import _mount_self_obj

###############################################################################
#   Datastore TestCase Implementation
###############################################################################
class DatastoreTestCase(unittest.TestCase):

    def setUp(self):
        self.sample_data = {
            'id': 'dats_12345678',
            'name': 'My Datastore',
            'description': 'This is a sample datastore.',
            'tags': ['tag1', 'tag2'],
            'link_uid': None,
            'workspace': 'my_workspace',
            'date': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'level': '1',
            'data': {'key1': 'value1', 'key2': 'value2'},
            'version': 'v1',
            'created_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'updated_on': DateTime(2023, 9, 9, 9, 9, 9, 9)
        }
        self.datastore = Datastore(**self.sample_data)

    def test_get_id_prefix(self):
        self.assertEqual(Datastore.get_id_prefix(), settings.DATASTORE_ID_PREFIX)

    def test_validate(self):
        datastore: Datastore = self.datastore.copy()
        with unittest.mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = True
            result = datastore.validate()

        mock_http_connection.assert_called_once_with(class_name='Datastore', params={}, method_name='validate', self_obj=_mount_self_obj(datastore))
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertTrue(result)

    def test_validate_error(self):
        datastore: Datastore = self.datastore.copy()
        datastore.data = DateTime.now()
        with self.assertRaisesRegex(FieldValueError, "Datastore data is not a valid json"):
            datastore.validate()

        datastore.data = None
        with self.assertRaisesRegex(RequiredError, "The data attribute is required"):
            datastore.validate()

    def test_to_dict(self):
        dict_output = self.datastore.to_dict()
        self.assertIsInstance(dict_output, dict)
        self.assertEqual(dict_output['name'], self.sample_data['name'])
        self.assertEqual(dict_output['date'], Date.strftime_or_null(self.sample_data['date']))
        self.assertEqual(dict_output['created'], self.sample_data['created_on'].timestamp())

    def test_to_dict_no_modify_obj_by_reference(self):
        datastore = Datastore(
            date=DateTime(2021, 1, 1, 1, 1, 1),
            created_on=DateTime(2021, 1, 1, 1, 1, 1),
            updated_on=DateTime(2021, 1, 1, 1, 1, 1),
            tags=['tag1', 'tag2'],
            data={'test_key': 'test_value'})
        dict_output = datastore.to_dict()

        self.assertDictEqual(
            dict_output,
            {
                'version': 'v1',
                'id': None,
                'name': None,
                'description': None,
                'tags': ['tag1',
                'tag2'],
                'link_uid': None,
                'workspace': None,
                'date': '20210101',
                'level': None,
                'data': {
                    'test_key': 'test_value'
                },
                'date_time': '20210101 01:01:01',
                'created': 1609462861,
                'updated': 1609462861
            }
        )

        dict_output['tags'].append('tag3')
        self.assertNotIn('tag3', datastore.tags)
        dict_output['data']['other_key'] = ['other_value']
        self.assertNotIn('other_key', datastore.data)

    def test_query_load_with_id(self):
        with unittest.mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = self.sample_data
            result = Datastore(id='dats_1234567891011211234567890', workspace='SampleWorkspace').load()

        mock_http_connection.assert_called_once_with(
            params={'entity_id': 'dats_1234567891011211234567890'},
            class_name='Datastore',
            method_name='retrieve',
            self_obj=None
        )
        mock_http_connection.return_value.get_response.assert_called_once_with()

        self.assertEqual(result, self.datastore)
        self.assertIsInstance(result, Datastore)

    def test_query_load(self):
        with unittest.mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = self.sample_data
            result = Datastore(link_uid='SampleLinkUID', workspace='SampleWorkspace').load()

        mock_http_connection.assert_called_once_with(
            self_obj=_mount_self_obj({
                'filters': [('workspace', '=', 'SampleWorkspace'), ('link_uid', '=', 'SampleLinkUID')],
                'order': [],
                'projection': None,
                'limit': None,
                'offset': None,
                'page_size': None,
                'page_token': None,
                'distinct_on': [],
                '_klass': 'Datastore',
                '_clean_order': []
            }),
            params={'offset': None},
            class_name='Query',
            method_name='load'
        )
        mock_http_connection.return_value.get_response.assert_called_once_with()

        self.assertEqual(result, self.datastore)
        self.assertIsInstance(result, Datastore)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from unittest import TestCase, mock
from unittest.mock import MagicMock

from everysk.sdk.entities.query import Query
from everysk.sdk.entities.portfolio.base import Portfolio

from everysk.sdk.base import _mount_self_obj

###############################################################################
#   Script TestCase Implementation
###############################################################################
class ScriptTestCase(TestCase):
    def setUp(self):
        self.mock_klass = MagicMock()
        self.query = Query(Portfolio)

    def test_script_query_none_user_input(self):
        result = Portfolio.script.fetch(None, 'someVariant', 'someWorkspace')
        self.assertIsNone(result)

    def test_script_query_previous_workers_without_id(self):
        user_input = {"someField": "someValue"}
        with mock.patch.object(Portfolio, '__init__', return_value=None):
            result = Portfolio.script.fetch(user_input, 'previousWorkers', 'someWorkspace')
        self.assertIsInstance(result, Portfolio)

    def test_script_query_tag_latest_variant(self):
        user_input = {"tags": "someTag"}

        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = Portfolio(name='test', workspace='someWorkspace')
            result = Portfolio.script.fetch(user_input, 'tagLatest', 'someWorkspace')
        self.assertIsInstance(result, Portfolio)
        self.assertEqual(result.name, 'test')

    def test_persist_transient(self):
        entity = Portfolio(name='test', workspace='someWorkspace')
        entity.validate = MagicMock()

        with mock.patch('everysk.sdk.entities.script.Script.get_response', return_value=entity) as mock_get_response:
            result = Portfolio.script.persist(entity, 'transient')

        mock_get_response.assert_called_once_with(self_obj=Portfolio.script, params={'entity': entity, 'persist': 'transient'})
        self.assertEqual(result, entity)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# Classes that handle HTTP Connections.

# Url: https://requests.readthedocs.io/en/latest/
#      https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers
#      https://developer.mozilla.org/en-US/docs/Glossary/Quality_values
#
import random
import time

import requests
from everysk.core.exceptions import HttpError, InvalidArgumentError
from everysk.core.object import BaseObject
from everysk.core.serialize import dumps, loads
from everysk.core.fields import BoolField, DictField, IntField, StrField
from everysk.core.log import Logger
from everysk.config import settings


log = Logger(name='everysk-lib-core-http-log')


class HttpConnectionConfig(BaseObject):
    # Activate/deactivate the use of the verify flag on HTTP requests.
    # By default this is defined in settings.HTTP_REQUESTS_VERIFY
    # but could be defined in the class configuration too.
    ssl_verify = BoolField(default=True)

    # Limit for retries
    retry_limit = IntField(default=5)

    # Times that randrange will use to do the next retry
    retry_end_seconds = IntField(default=30)
    retry_start_seconds = IntField(default=5)

    def get_ssl_verify(self) -> bool:
        return self.ssl_verify if settings.HTTP_REQUESTS_VERIFY is Undefined else settings.HTTP_REQUESTS_VERIFY


class HttpConnection(BaseObject):
    """
    Base class to use for HTTP connections, has two attributes:
        - timeout: It's in and represent seconds, defaults to 30.
        - url: It's string and will be the destination.
    """
    class Config(HttpConnectionConfig):
        pass

    ## Private attributes
    _retry_count = IntField(default=1) # Used to control how many times this connection was retry

    ## Public attributes
    config: Config # To autocomplete correctly
    headers = DictField(default=None)
    timeout = IntField(default=30)
    url = StrField(default=None)

    def _clean_response(self, response: requests.models.Response) -> requests.models.Response:
        """
        Checks status_code for response, if status_code is different than 200 throws an exception.

        Args:
            response (requests.models.Response): Http response from server.

        Raises:
            HttpError: If something goes wrong raise exception with status_code and content.
        """
        if getattr(response, 'status_code', settings.HTTP_SUCCESS_STATUS_CODES[0]) not in settings.HTTP_SUCCESS_STATUS_CODES:
            raise HttpError(status_code=response.status_code, msg=response.content)

        return response

    def get_headers(self) -> dict:
        """
        Headers needed to send HTTP methods.
        Below are the most common Headers used by browsers,
        we use them to look less like a Bot and more like a valid access.
        """
        headers = settings.HTTP_DEFAULT_HEADERS.copy()
        if self.headers is not None:
            headers.update(self.headers)

        return headers

    def get_url(self) -> str:
        """
        Generate the correct url to fetch data from vendor on POST/GET requests.
        """
        return self.url

    def message_error_check(self, message: str) -> bool: # pylint: disable=unused-argument
        """
        If this method returns True, the connection will be tried again.

        Args:
            message (str): The error message that occurred on the connection.
        """
        return False

    def _get_response_from_url(self) -> requests.models.Response:
        """
        This will be implemented on child classes to really do the connection.
        """
        return None

    def get_response(self) -> requests.models.Response:
        """
        Try to fetch data from self.get_url and calling self._get_response_from_url for the complete response.
        On HttpError, if self.message_error_check is True we will try connect again for a few more times.
        """
        try:
            response = self._clean_response(self._get_response_from_url())
            # After a success we set the value to 1 again
            self._retry_count = 1
        except Exception as error: # pylint: disable=broad-exception-caught
            # Sometimes it can happen that the server is busy, if this happen the error message must be tested
            # and must return true to enable recursion and we will try again the connection.
            if self.message_error_check(str(error).lower()) and self._retry_count < self.config.retry_limit:
                self._retry_count += 1
                # As we have several processes, we use a random number to avoid collision between them.
                time.sleep(random.randint(self.config.retry_start_seconds, self.config.retry_end_seconds))
                response = self.get_response()
            else:
                raise error

        return response


class HttpGETConnection(HttpConnection):
    """ Class that implements a interface for HTTP GET connections """
    params = DictField()
    user = StrField()
    password = StrField()

    def get_params(self) -> dict:
        """
        This method is used to make the correct params to pass on GET request.
        These params will be added to the URL with & separating then.
        """
        return self.params

    def _get_response_from_url(self) -> requests.models.Response:
        """
        Try to fetch data from url using GET request.
        Note that any dictionary key whose value is None will not be added to the URL's query string.
        """
        params = {
            'url': self.get_url(),
            'headers': self.get_headers(),
            'params': self.get_params(),
            'verify': self.config.get_ssl_verify(),
            'timeout': self.timeout
        }
        if self.user:
            params['auth'] = (self.user, self.password)

        if settings.HTTP_LOG_RESPONSE:
            dct = params.copy()
            # To remove the password in the logs
            if 'auth' in params:
                dct['auth'] = (params['auth'][0], '***********')

            log.debug('HTTP GET request: %s', dct)

        response = requests.get(**params)

        if settings.HTTP_LOG_RESPONSE:
            dct = {
                'status_code': response.status_code,
                'time': response.elapsed.total_seconds(),
                'headers': response.headers,
                'content': response.content
            }
            log.debug('HTTP GET response: %s', dct)

        return response


class HttpPOSTConnection(HttpConnection):
    """
    Class that implements a interface for HTTP POST connections.
    If self.is_json is True the POST method will be a JSON POST,
    otherwise will be a Form POST Data.
    """
    is_json = BoolField(default=True)
    payload = DictField()

    def get_headers(self) -> dict:
        """ Headers needed to send HTTP Post methods. """
        headers = super().get_headers()
        if self.is_json:
            headers['Content-Type'] = 'application/json; charset=utf-8'
        else:
            headers['Content-Type'] = 'application/x-www-form-urlencoded'

        return headers

    def get_payload(self) -> dict:
        """ Make the correct payload body to pass on POST request. """
        return self.payload

    def _get_response_from_url(self) -> requests.models.Response:
        """ Try to get/set data on url using POST request. """
        params = {
            'url': self.get_url(),
            'headers': self.get_headers(),
            'verify': self.config.get_ssl_verify(),
            'timeout': self.timeout
        }
        if self.is_json:
            params['json'] = self.get_payload()
        else:
            params['data'] = self.get_payload()

        if settings.HTTP_LOG_RESPONSE:
            log.debug('HTTP POST request: %s', params)

        response = requests.post(**params)

        if settings.HTTP_LOG_RESPONSE:
            dct = {
                'status_code': response.status_code,
                'time': response.elapsed.total_seconds(),
                'headers': response.headers,
                'content': response.content
            }
            log.debug('HTTP POST response: %s', dct)

        return response


class HttpPOSTCompressedConnection(HttpPOSTConnection):

    def get_payload(self) -> dict:
        """ Make the correct payload body to pass on POST request. """
        return dumps(self.payload)

    def get_response(self) -> dict:
        """
        Try to fetch data from self.get_url and calling self._get_response_from_url for the complete response.
        On HttpError, if self.message_error_check is True we will try connect again more 5 times.
        Decompress the response.content
        """
        response = super().get_response()
        return loads(response.content)


class HttpSDKPOSTConnection(HttpPOSTCompressedConnection):

    is_json = BoolField(default=False, readonly=True)

    class_name = StrField()
    method_name = StrField()
    self_obj = DictField()
    params = DictField()

    def get_url(self) -> str:
        return f'{settings.EVERYSK_API_URL}/{settings.EVERYSK_SDK_VERSION}/{settings.EVERYSK_SDK_ROUTE}'

    def get_payload(self) -> dict:
        """ Make the correct payload body to pass on POST request. """
        self.payload = {
            'class_name': self.class_name,
            'method_name': self.method_name,
            'self_obj': self.self_obj,
            'params': self.params
        }
        return super().get_payload()

    def get_headers(self) -> dict:
        """ Headers needed to send HTTP Post methods. """
        headers = super().get_headers()
        everysk_api_sid = settings.EVERYSK_API_SID
        everysk_api_token = settings.EVERYSK_API_TOKEN

        if not everysk_api_sid:
            raise InvalidArgumentError('Invalid API SID')
        if not everysk_api_token:
            raise InvalidArgumentError('Invalid API TOKEN')

        headers['Authorization'] = f'Bearer {everysk_api_sid}:{everysk_api_token}'

        return headers



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from unittest import TestCase, mock

from everysk.config import settings

from everysk.core.datetime import DateTime, Date

from everysk.sdk.entities.portfolio.base import SecuritiesField, Portfolio
from everysk.sdk.entities.portfolio.securities import Securities
from everysk.sdk.entities.portfolio.security import Security

from everysk.sdk.base import _mount_self_obj

###############################################################################
#   Securities Field TestCase Implementation
###############################################################################
class TestSecuritiesField(TestCase):

    def setUp(self):
        self.securities_field = SecuritiesField()

    def test_clean_value_with_list(self):
        test_value = [{"symbol": "AAPL", "name": "Apple Inc."}]
        cleaned_value = self.securities_field.clean_value(test_value)

        # Ensure the cleaned value is of type `Securities`
        self.assertIsInstance(cleaned_value, Securities)

    def test_clean_value_without_list(self):
        test_value = "Some value"
        cleaned_value = self.securities_field.clean_value(test_value)

        # Ensure the cleaned value is not modified
        self.assertEqual(cleaned_value, [test_value])

###############################################################################
#   Portfolio TestCase Implementation
###############################################################################
class TestPortfolio(TestCase):

    def setUp(self):
        self.sample_data = {
            'id': 'port_1234567891011211234567890',
            'workspace': 'SampleWorkspace',
            'name': 'SamplePortfolio',
            'tags': ['tag1', 'tag2'],
            'link_uid': 'ABC',
            'description': 'Description',
            'nlv': 1000.0,
            'base_currency': 'USD',
            'date': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'securities': Securities([{'symbol': 'AAPL'}]),
            'version': '1.0',
            'created_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'updated_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'level': 'v1',
            'outstanding_shares': 1000,
            'source_hash': 'XYZ',
            'status': 'OK',
            'portfolio_uid': 'UID123',
            'check_securities': False
        }
        self.portfolio = Portfolio(**self.sample_data)

    def test_static_methods(self):
        self.assertEqual(Portfolio.get_id_prefix(), settings.PORTFOLIO_ID_PREFIX)

    def test_validate(self):
        portfolio: Portfolio = self.portfolio.copy()
        portfolio.check_securities = True
        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = True
            result = portfolio.validate()

        mock_http_connection.assert_called_once_with(class_name='Portfolio', params={}, method_name='validate', self_obj=_mount_self_obj(portfolio))
        mock_http_connection.return_value.get_response.assert_called_once_with()
        self.assertTrue(result)

    def test_to_dict(self):
        dict_output = self.portfolio.to_dict()
        self.assertIsInstance(dict_output, dict)
        self.assertEqual(dict_output['name'], self.sample_data['name'])
        self.assertEqual(dict_output['date'], Date.strftime_or_null(self.sample_data['date']))
        self.assertEqual(dict_output['created'], self.sample_data['created_on'].timestamp())

    def test_check(self):
        entity_dict = self.sample_data
        entity_dict['check_securities'] = True
        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = self.sample_data
            result = Portfolio.check(entity_dict=entity_dict)

        mock_http_connection.assert_any_call(class_name='Portfolio', params={}, method_name='validate', self_obj=_mount_self_obj(Portfolio(**entity_dict)))
        mock_http_connection.return_value.get_response.assert_called_once_with()

        check_port = self.portfolio.copy()
        check_port.check_securities = True
        self.assertEqual(result, check_port)
        self.assertIsInstance(result, Portfolio)

    def test_to_csv(self):
        out_csv = self.portfolio.to_csv()
        self.assertIsInstance(out_csv, str)

    def test_to_dict_without_internals(self):
        dict_output = self.portfolio.to_dict(with_internals=False)
        self.assertIsInstance(dict_output, dict)
        self.assertEqual(dict_output['securities'][0], Security(symbol='AAPL').to_dict(with_internals=False))

    def test_to_dict_no_modify_obj_by_reference(self):
        portfolio = Portfolio(
            date=DateTime(2021, 1, 1, 1, 1, 1),
            created_on=DateTime(2021, 1, 1, 1, 1, 1),
            updated_on=DateTime(2021, 1, 1, 1, 1, 1),
            tags=['tag1', 'tag2'],
            securities=[{
                'symbol': 'AAPL',
                'maturation_date': Date(2021, 1, 1),
                'extra_data': {
                    'date': Date(2021, 1, 1),
                    'datetime': DateTime(2021, 1, 1, 1, 1, 1),
                    'date_list': [Date(2021, 1, 1)],
                    'datetime_list': [DateTime(2021, 1, 1, 1, 1, 1)],
                    'dict': {'a': 1, 'b': 2},
                    'tuple_list': [(Date(2021, 1, 1),)],
                    'date_tuple': (Date(2021, 1, 1),),
                    'date_set': set([Date(2021, 1, 1)])
                }}])
        dict_output = portfolio.to_dict(with_internals=False)

        self.assertDictEqual(
            dict_output,
            {
                'id': None,
                'workspace': None,
                'name': None,
                'tags': ['tag1', 'tag2'],
                'link_uid': None,
                'description': None,
                'nlv': None,
                'base_currency': None,
                'date': '20210101',
                'securities': [
                    {
                        'status': None,
                        'id': None,
                        'symbol': 'AAPL',
                        'quantity': None,
                        'instrument_class': None,
                        'ticker': None,
                        'label': None,
                        'name': None,
                        'isin': None,
                        'exchange': None,
                        'currency': None,
                        'fx_rate': None,
                        'market_price': None,
                        'premium': None,
                        'market_value': None,
                        'market_value_in_base': None,
                        'instrument_type': None,
                        'instrument_subtype': None,
                        'asset_class': None,
                        'asset_subclass': None,
                        'error_type': None,
                        'error_message': None,
                        'maturity_date': None,
                        'indexer': None,
                        'percent_index': None,
                        'rate': None,
                        'coupon': None,
                        'multiplier': None,
                        'underlying': None,
                        'series': None,
                        'option_type': None,
                        'strike': None,
                        'issue_price': None,
                        'issue_date': None,
                        'issuer': None,
                        'issuer_type': None,
                        'cost_price': None,
                        'unrealized_pl': None,
                        'unrealized_pl_in_base': None,
                        'book': None,
                        'trader': None,
                        'trade_id': None,
                        'operation': None,
                        'accounting': None,
                        'warranty': None,
                        'return_date': None,
                        'settlement': None,
                        'look_through_reference': None,
                        'extra_data': {
                            'maturation_date': '20210101',
                            'date': '20210101',
                            'datetime': '20210101 01:01:01',
                            'date_list': ['20210101'],
                            'datetime_list': ['20210101 01:01:01'],
                            'dict': {'a': 1, 'b': 2},
                            'tuple_list': [('20210101',)],
                            'date_tuple': ('20210101',),
                            'date_set': ['20210101']
                        },
                        'hash': None
                    }
                ],
                'version': 'v1',
                'level': None,
                'check_securities': False,
                'date_time': '20210101 01:01:01',
                'created': 1609462861,
                'updated': 1609462861
            }
        )

        dict_output['tags'].append('tag3')
        self.assertNotIn('tag3', portfolio.tags)

        dict_output['securities'][0]['extra_data']['date_list'].append('20210102')
        self.assertNotIn('20210102', portfolio.securities[0].extra_data.date_list)

        dict_output['securities'][0]['extra_data']['datetime_list'].append('20210102 01:01:01')
        self.assertNotIn('20210102 01:01:01', portfolio.securities[0].extra_data.datetime_list)

        dict_output['securities'][0]['extra_data']['tuple_list'].append(('20210102',))
        self.assertNotIn(('20210102',), portfolio.securities[0].extra_data.tuple_list)

        dict_output['securities'][0]['extra_data']['date_tuple'] += ('20210102',)
        self.assertNotIn('20210102', portfolio.securities[0].extra_data.date_tuple)

        dict_output['securities'][0]['extra_data']['date_set'].append('20210102')
        self.assertNotIn('20210102', portfolio.securities[0].extra_data.date_set)

        dict_output['securities'][0]['extra_data']['dict']['c'] = 3
        self.assertNotIn('c', portfolio.securities[0].extra_data.dict)

    def test_query_load_with_id(self):
        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = self.sample_data
            result = Portfolio(id='port_1234567891011211234567890', workspace='SampleWorkspace').load()

        mock_http_connection.assert_called_once_with(
            params={'entity_id': 'port_1234567891011211234567890'},
            class_name='Portfolio',
            method_name='retrieve',
            self_obj=None
        )
        mock_http_connection.return_value.get_response.assert_called_once_with()

        self.assertEqual(result, self.portfolio)
        self.assertIsInstance(result, Portfolio)

    def test_query_load(self):
        with mock.patch('everysk.sdk.base.HttpSDKPOSTConnection') as mock_http_connection:
            mock_http_connection.return_value.get_response.return_value = self.sample_data
            result = Portfolio(link_uid='SampleLinkUID', workspace='SampleWorkspace').load()

        mock_http_connection.assert_called_once_with(
            self_obj=_mount_self_obj({
                'filters': [('workspace', '=', 'SampleWorkspace'), ('link_uid', '=', 'SampleLinkUID')],
                'order': [],
                'projection': None,
                'limit': None,
                'offset': None,
                'page_size': None,
                'page_token': None,
                'distinct_on': [],
                '_klass': 'Portfolio',
                '_clean_order': []
            }),
            params={'offset': None},
            class_name='Query',
            method_name='load'
        )
        mock_http_connection.return_value.get_response.assert_called_once_with()

        self.assertEqual(result, self.portfolio)
        self.assertIsInstance(result, Portfolio)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from typing import Any
import inspect
import requests

from everysk.core.string import import_from_string
from everysk.core.object import BaseDict
from everysk.core.exceptions import HttpError, SDKError
from everysk.core.http import HttpSDKPOSTConnection

###############################################################################
#   Base Entity Implementation
###############################################################################
def _parser_out(obj: Any, with_internals: bool = True) -> Any:
    """
    Recursively parse an object and its nested data structures.

    This function recursively processes the input object and its nested data structures.
    It can be used to convert objects with custom 'to_dict' and 'strftime_or_null' methods
    into dictionaries while handling common data types like dictionaries, lists, tuples, and sets.

    Args:
        obj (Any): The object to be parsed.
        with_internals (bool, optional): Flag to include internal details when parsing.
            Default is True.

    Returns:
        Any: The parsed object.

    """
    ret: Any = obj
    if hasattr(obj, 'to_dict'):
        ret = obj.to_dict(with_internals)

    elif isinstance(obj, dict):
        ret = type(obj)()
        for key, value in obj.items():
            ret[key] = _parser_out(value, with_internals)

    elif isinstance(obj, list):
        ret = type(obj)()
        for item in obj:
            ret.append(_parser_out(item, with_internals))

    elif isinstance(obj, tuple):
        ret = []
        for item in obj:
            ret.append(_parser_out(item, with_internals))
        ret = type(obj)(ret)

    elif isinstance(obj, set):
        ret = []
        for item in obj:
            ret.append(_parser_out(item, with_internals))

    elif obj == Undefined:
        ret = Undefined.default_parse_string

    elif hasattr(obj, 'strftime_or_null'):
        ret = obj.__class__.strftime_or_null(obj)

    return ret

def _mount_self_obj(self_obj: Any) -> dict | None:
    """
    This function mounts the self_obj with there private attributes and
    returns the dictionary.

    Args:
        self_obj (Any): The object to be mounted.

    Returns:
        dict: The parsed object. If the object is not serializable, returns None.
    """
    if self_obj is None:
        return None

    ret = dict(self_obj)

    if hasattr(self_obj, '__dict__'):
        for key, value in self_obj.__dict__.items():
            if key.startswith('_') and not inspect.ismethod(value):
                if inspect.isclass(value):
                    ret[key] = value.__name__
                elif isinstance(value, BaseDict):
                    ret[key] = _mount_self_obj(value)
                elif isinstance(value, (tuple, set)):
                    ret[key] = list(value)
                else:
                    ret[key] = value

    return ret

class BaseSDK(BaseDict):
    """
    A base class for SDK classes.

    This class provides a base implementation for SDK classes.
    """
    # This is a blacklist to not create keys
    _keys_blacklist: frozenset[str] = frozenset([
        'config',
        'query',
        'script'
    ])

    @classmethod
    def get_response(cls, **kwargs: dict) -> Any:
        """
        Get a response from an SDK method.

        This method sends an HTTP POST request to a remote service and returns the response as a dictionary.

        Args:
            **kwargs (dict): Keyword arguments used to configure the HTTP request and SDK behavior.

        Keyword Args:
            class_name (str, optional): The name of the SDK class making the request.
                Defaults to the class name of the calling class.
            method_name (str, optional): The name of the SDK method making the request.
                Defaults to the name of the calling function.
            self_obj (object, optional): An instance of the calling SDK class, if applicable. Defaults to None.
            params (dict, optional): Additional parameters to include in the HTTP request. Defaults to an empty dictionary.

        Returns:
            Any: The response from the remote service.

        Raises:
            SDKError: If there is an issue with the SDK operation.
        """

        # Set default values for keyword arguments if not provided
        kwargs.setdefault('class_name', cls.__name__)
        kwargs.setdefault('method_name', inspect.stack()[1].function)
        kwargs.setdefault('self_obj', None)
        kwargs.setdefault('params', {})

        kwargs['self_obj'] = _mount_self_obj(kwargs['self_obj'])

        try:
            response: requests.models.Response = HttpSDKPOSTConnection(**kwargs).get_response()
        except HttpError as error:
            raise SDKError(error.msg) from error

        if response and isinstance(response, dict) and 'error_message' in response and 'error_module' in response:
            error_module = import_from_string(response['error_module'])
            raise error_module(response['error_message'])

        return response

    def to_dict(self, with_internals: bool = True) -> dict: # pylint: disable=unused-argument
        """
        Convert the object to a dictionary.

        This method converts the object to a dictionary by recursively parsing its attributes.

        Args:
            with_internals (bool, optional): Flag to include internal details when parsing.
                Default is True.

        Returns:
            dict: The object as a dictionary.

        Example usage:
            To convert an object to a dictionary:
            >>> obj = MyObject()
            >>> obj_dict = obj.to_dict()
        """
        return _parser_out(BaseDict(**self.copy()), with_internals)
